import unittest
from lxml import etree as et
from xml_common.utils import Language
from eml.resources import EMLKeywordSet, KeywordType
from test_xml.test_xml import TestXML


class TestKeyword(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng"
    }

    def test_parse_simple(self):
        text_xml = """
<keywordSet>
    <keyword>taxonomic</keyword>
    <keyword>database</keyword>
    <keyword>online</keyword>
    <keyword>species</keyword>
    <keywordThesaurus/>
</keywordSet>
        """
        keyword = EMLKeywordSet.from_string(text_xml)
        self.assertEqual(4, len(keyword.keywords), "Error on parsing keywords")
        self.assertEqual("", keyword.thesaurus, "Error on parsing thesaurus")
        self.assertEqual(4, len(keyword.keywords_type), "Error on parsing keywords")
        self.assertEqual([KeywordType.NULL] * 4, keyword.keywords_type, "Error on type of keywords")
        self.assertEqualTree(et.fromstring(text_xml), keyword.to_element(), "Error on to element")

    def test_parse_complete(self):
        text_xml = """
<keywordSet>
    <keyword keywordType="place" xml:lang="esp">Concepción</keyword>
    <keyword>database</keyword>
    <keyword>online</keyword>
    <keyword>species</keyword>
    <keyword xml:lang="esp">especies</keyword>
    <keyword keywordType="taxonomic">plantae</keyword>
    <keywordThesaurus>IRIS keyword thesaurus</keywordThesaurus>
</keywordSet>
        """
        keyword = EMLKeywordSet.from_string(text_xml)
        self.assertEqual(6, len(keyword.keywords), "Error on parsing keywords")
        self.assertEqual("IRIS keyword thesaurus", keyword.thesaurus, "Error on parsing thesaurus")
        self.assertEqual(6, len(keyword.keywords_type), "Error on parsing keywords")
        self.assertEqual(
            [
                KeywordType.PLACE,
                KeywordType.NULL,
                KeywordType.NULL,
                KeywordType.NULL,
                KeywordType.NULL,
                KeywordType.TAXONOMIC,
            ], keyword.keywords_type,
            "Error on type of keywords"
        )
        self.assertEqual(Language.ESP, keyword.keywords[4].language, "Error on keyword language")
        self.assertEqualTree(et.fromstring(text_xml), keyword.to_element(), "Error on to element")

    def test_keyword_type_default(self):
        keyword = EMLKeywordSet(["species", "database", "online", "taxonomic"])
        self.assertEqual(4, len(keyword.keywords_type), "Error on default types")
        self.assertEqual([KeywordType.NULL] * 4, keyword.keywords_type, "Error on default types specific")

    def test_keyword_language_default(self):
        keyword = EMLKeywordSet(["species", "database", "online", "taxonomic"])
        self.assertEqual(4, len(keyword.keywords), "Error on keywords")
        self.assertEqual([Language.ENG] * 4, [k.language for k in keyword.keywords], "Error on default languages")

    def test_keyword_invalid(self):
        self.assertRaises(
            ValueError,
            EMLKeywordSet,
            []
        )

    def test_keyword_invalid_parse(self):
        text_xml = """
<keywordSet>
</keywordSet>
        """
        self.assertRaises(
            ValueError,
            EMLKeywordSet.from_string,
            text_xml,
        )
        text_xml = """
<keywordSet>
    <keywordThesaurus>Example Thesaurus</keywordThesaurus>
</keywordSet>
        """
        self.assertRaises(
            ValueError,
            EMLKeywordSet.from_string,
            text_xml,
        )

    def test_keyword_type(self):
        keyword = EMLKeywordSet(
            ["conservation", "database", "online", "taxonomic"],
            keywords_type=[KeywordType.THEME]
        )
        self.assertEqual(4, len(keyword.keywords_type), "Error on default types")
        self.assertEqual([
            KeywordType.THEME,
            KeywordType.NULL,
            KeywordType.NULL,
            KeywordType.NULL,
        ], keyword.keywords_type,
            "Error on default types specific"
        )

    def test_parse_none(self):
        self.assertIsNone(EMLKeywordSet.parse(None, {}), "Parse from nowhere")

    def test_parse_invalid_type(self):
        text_xml = """
<keywordSet>
    <keyword keywordType="place" xml:lang="esp">Concepción</keyword>
    <keyword>database</keyword>
    <keyword>online</keyword>
    <keyword keywordType="species">species</keyword>
    <keyword keywordType="taxonomic">plantae</keyword>
    <keywordThesaurus>IRIS keyword thesaurus</keywordThesaurus>
</keywordSet>
        """
        keyword = EMLKeywordSet.from_string(text_xml)
        self.assertEqual(5, len(keyword.keywords), "Error on parsing keywords")
        self.assertEqual("IRIS keyword thesaurus", keyword.thesaurus, "Error on parsing thesaurus")
        self.assertEqual(5, len(keyword.keywords_type), "Error on parsing keywords")
        self.assertEqual(
            [
                KeywordType.PLACE,
                KeywordType.NULL,
                KeywordType.NULL,
                KeywordType.NULL,
                KeywordType.TAXONOMIC,
            ], keyword.keywords_type,
            "Error on type of keywords"
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Any

from lxml import etree as et
from enum import Enum

from xml_common import XMLObject
from eml.types import ExtensionString


class Scope(Enum):
    DOCUMENT = 0
    SYSTEM = 1


class EMLObject(XMLObject, ABC):
    """
    Abstract class representing most of the EML document

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    """
    def __init__(
            self, _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None
    ) -> None:
        super().__init__()
        self.__id__ = _id
        self.__scope__ = scope
        self.__system__ = system
        self.__referencing__ = referencing
        if self.__referencing__:
            assert self.__id__ is not None, "Must give the id of the element being referenced."
        self.__references__ = ExtensionString(self.__id__, system=references_system)
        return

    @property
    def id(self) -> str:
        """str: Unique identifier of the resource"""
        return self.__id__

    @property
    def scope(self) -> Scope:
        """Scope: The scope of the identifier."""
        return self.__scope__

    @property
    def system(self) -> str:
        """str: The data management system within which an identifier is in scope and therefore unique."""
        return self.__system__

    @property
    def referencing(self) -> bool:
        """bool: Whether the resource is referencing another or is being defined"""
        return self.__referencing__

    @property
    def references(self) -> ExtensionString:
        """bool: Whether the resource is referencing another or is being defined"""
        return self.__references__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLObject | None:
        """
        Common parse of an EML object

        Parameters
        ----------
        element : lxml.etree.Element
            XML Element to parse
        nmap : Dict
            Namespace

        Returns
        -------
        EMLObject
            Parsed EML Object
        """
        if element is None:
            return None
        _id = element.get("id", None)
        references = element.find("references", nmap)
        if _id is not None and references is not None:
            raise AssertionError("""
If an element references another element, it may not have an id. This prevents circular references.
See https://eml.ecoinformatics.org/validation-and-content-references#id-and-scope-examples
            """)
        if references is not None:
            eml_object = cls.get_referrer(element, nmap)
        else:
            eml_object = cls.get_no_referrer(element, nmap)
        eml_object.__namespace__ = nmap
        return eml_object

    @classmethod
    @abstractmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLObject:
        """
        Generate an EML Object referencing another EML Object.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLObject
            Object parsed that reference another.
        """
        pass

    @classmethod
    @abstractmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLObject:
        """
        Generate an EML Object that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLObject
            Object parsed.
        """
        pass

    @staticmethod
    def get_scope(element: et.Element) -> Scope:
        """
        Get the scope from the element.

        Parameters
        ----------
        element : lxml.etree.Element
            XML Element to find scope.

        Returns
        -------
        Scope
            A valid Scope.

        Raises
        ------
        ValueError
            Not a valid scope string.
        """
        scope = element.get("scope", None)
        if scope is None or scope == "document":
            return Scope.DOCUMENT
        elif scope == "system":
            return Scope.SYSTEM
        else:
            raise ValueError(f"{scope} is not a valid scope ('document' or 'system')")

    def generate_references_element(self) -> et.Element:
        """
        Generate the `<references>` element.

        Returns
        -------
        xml.etree.Element
            XML `Element` representing the references of referrer resource.
        """
        if not self.referencing:
            return None
        else:
            self.__references__.set_tag("references")
            return self.__references__.to_element()

    def _to_element_(self, element: et.Element) -> et.Element:
        """
        Add references values and xml_common values to element.

        Returns
        -------
        lxml.tree.Element
            Object in the Element format.
        """
        references = self.generate_references_element()
        if references is not None:
            element.append(references)
        else:
            if self.id is not None:
                element.set("id", self.id)
        element.set("scope", self.scope.name.lower())
        if self.system is not None:
            element.set("system", self.system)
        return element

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EMLObject):
            return (
                    self.id == other.id and
                    self.scope == other.scope and
                    self.system == other.system and
                    self.referencing == other.referencing and
                    self.references == other.references
            )
        else:
            return False



import datetime
import os
import unittest

import requests
import datetime as dt
from lxml import etree as et

from eml import EML, EMLVersion
from xml_common.utils import Language
from eml.resources import EMLResource, EMLKeywordSet, EMLLicense, EMLDistribution, EMLCoverage
from eml.resources.coverage import TemporalCoverage
from eml.resources.distribution import EMLOnline, EMLOffline
from eml.types import Scope, ResponsibleParty, IndividualName, OrganizationName, PositionName, Role, I18nString, \
    EMLTextType, SemanticAnnotation
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestEML(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
        "phonetype": "voice",
    }

    def setUp(self) -> None:
        eml_path = os.path.join(PATH, os.pardir, "example_data", "eml.xml")
        self.eml = EML.from_xml(eml_path)
        with open(eml_path, "r", encoding="utf-8") as file:
            self.text_eml = file.read()
        self.element_eml = et.fromstring(self.text_eml)
        self.empty_eml = EML(
            package_id="Example package",
            system="http://my.system",
            resource_type=EMLResource.DATASET,
        )
        return

    def initialize_resource(self, **kwargs):
        self.empty_eml.initialize_resource(
            "Example Title",
            ResponsibleParty(individual_name=IndividualName(first_name="Joe", last_name="Doe")),
            contact=[ResponsibleParty(position_name=PositionName("Contact"))],
            **kwargs
        )

    @staticmethod
    def check_online_test(text: str) -> bool:
        try:
            status = text.split(": ")[-1].lower().strip(".")
            if status == "passed":
                return True
            elif status == "failed":
                return False
            else:
                raise AssertionError(f"Unexpected status: {status}")
        except Exception as e:
            raise AssertionError(f"Error parsing response status: {e}\n\tParsing text:\n\t{text}")

    @staticmethod
    def validate(eml: str):
        url = "https://knb.ecoinformatics.org/emlparser/parse"
        form_data = {
            "action": "textparse",
            "doctext": eml
        }
        response = requests.post(url, data=form_data)
        if response.status_code == 200:
            root = et.fromstring(response.content, parser=et.HTMLParser())
            h4_elements = root.xpath('//h4')
            # EML Test
            eml_response = TestEML.check_online_test(h4_elements[0].text)
            if not eml_response:
                errors = "\n\t".join([error.text for error in h4_elements[0].getnext().getnext()])
                raise AssertionError(f"Error on parsing EML specific test:\n\t{errors}")
            # XML Test
            xml_response = TestEML.check_online_test(h4_elements[1].text)
            if not xml_response:
                raise AssertionError(f"Error on parsing XML specific test: {h4_elements[1].getnext().getnext().text}")
        else:
            raise AssertionError("Request failed with status code:", response.status_code)

    def test_simple_eml(self):
        self.initialize_resource()
        self.validate(self.empty_eml.to_xml())

    def test_parse(self):
        self.assertEqual("Example Package", self.eml.package_id, "Error on package id read")
        self.assertEqual("http://gbif.org", self.eml.system, "Error on system")
        self.assertEqual(Scope.SYSTEM, self.eml.scope, "Error on Scope")
        self.assertEqual(Language.ENG, self.eml.language, "Wrong language")
        self.assertEqual(EMLResource.DATASET, self.eml.resource_type, "Wrong resource type found")
        self.assertIsNone(self.eml.access, "Access from nothing")
        self.assertEqual(0, len(self.eml.annotations), "Annotations from nothing")
        self.assertEqual(1, len(self.eml.additional_metadata), "Additional metadata not found")
        self.assertEqualTree(self.element_eml, self.eml.to_element(), "Error on generated element")

    def test_eml_version(self):
        self.assertEqual(EMLVersion.LATEST, EMLVersion.get_version(None), "Wrong default")
        self.assertRaises(
            NotImplementedError,
            EMLVersion.get_version,
            "any_url another_url"
        )

    def test_str(self):
        self.assertEqual(
            """EML:
\tResource Type: DATASET
\tTitle: Example for Darwin Core Archive
\tCreator: Creator Organization
\tMetadataProvider: Metadata Manager at Metadata Provider Organization
\tCustodian Steward: Doe, J. (Custodian Steward)
\tOriginator: Doe, J.
\tAuthor: Example, J.""",
            str(self.eml),
            "Wrong string of EML"
        )

    def test_referrer(self):
        self.assertRaises(ValueError, self.eml.get_referrer, et.Element("example"), {})

    def test_set_title(self):
        self.initialize_resource()
        self.assertEqual(
            "Example Title",
            self.empty_eml.resource.title,
            "Title did not set"
        )
        self.assertEqual(
            Language.ENG,
            self.empty_eml.resource.title.language,
            "Language did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_title_lang(self):
        self.initialize_resource()
        self.empty_eml.add_title("Título de ejemplo", Language.ESP)
        self.assertEqual(
            "Título de ejemplo",
            self.empty_eml.resource.titles[1],
            "Title did not set"
        )
        self.assertEqual(
            Language.ESP,
            self.empty_eml.resource.titles[1].language,
            "Language did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_titles(self):
        self.empty_eml.initialize_resource(
            ["Example Title", I18nString("Título de ejemplo", Language.ESP)],
            ResponsibleParty(individual_name=IndividualName(first_name="Joe", last_name="Doe")),
            contact=[ResponsibleParty(position_name=PositionName("Contact"))]
        )
        self.assertEqual(
            "Example Title",
            self.empty_eml.resource.title,
            "Title did not set"
        )
        self.assertEqual(
            Language.ENG,
            self.empty_eml.resource.title.language,
            "Language did not set"
        )
        self.assertEqual(
            "Título de ejemplo",
            self.empty_eml.resource.titles[1],
            "Title did not set"
        )
        self.assertEqual(
            Language.ESP,
            self.empty_eml.resource.titles[1].language,
            "Language did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_creator(self):
        self.initialize_resource()
        self.assertEqual(
            "Doe, J.",
            str(self.empty_eml.resource.creator),
            "First creator did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_add_creator(self):
        self.empty_eml.initialize_resource(
            "Example title", [
                ResponsibleParty(individual_name=IndividualName(first_name="Joe", last_name="Doe")),
                ResponsibleParty(organization_name=OrganizationName("Example Organization")),
            ],
            contact=[ResponsibleParty(position_name=PositionName("Contact"))]
        )
        self.empty_eml.add_creator(
            ResponsibleParty(individual_name=IndividualName(first_name="John", last_name="Doe"))
        )
        self.assertEqual(
            "Doe, J.",
            str(self.empty_eml.resource.creator),
            "First creator did not set"
        )
        self.assertEqual(
            "Example Organization",
            str(self.empty_eml.resource.creators[1]),
            "Creator did not set"
        )
        self.assertEqual(
            "Doe, J.",
            str(self.empty_eml.resource.creators[2]),
            "Creator did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_alternative(self):
        self.initialize_resource()
        self.assertEqual(
            0, len(self.empty_eml.resource.alternative_identifiers),
            "Alternative identifiers initialized from nowhere"
        )
        self.empty_eml.add_alternative_identifier("Example Alternative Identifier")
        self.assertEqual(
            "Example Alternative Identifier",
            self.empty_eml.resource.alternative_identifiers[0],
            "Alternative identifier did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_short_name(self):
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.short_name)
        self.empty_eml.set_short_name("A short Name")
        self.assertEqual(
            "A short Name", self.empty_eml.resource.short_name,
            "Short Name did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_metadata_provider(self):
        self.initialize_resource()
        self.assertEqual(
            0, len(self.empty_eml.resource.metadata_provider),
            "Metadata Provider from nowhere"
        )
        self.empty_eml.add_metadata_provider(ResponsibleParty(
            individual_name=IndividualName(first_name="Joe", last_name="Doe")
        ))
        self.empty_eml.add_metadata_provider(ResponsibleParty(
            organization_name=OrganizationName("Example Organization")
        ))
        self.assertEqual(
            "Doe, J.",
            str(self.empty_eml.resource.metadata_provider[0]),
            "Metadata provider did not set"
        )
        self.assertEqual(
            "Example Organization",
            str(self.empty_eml.resource.metadata_provider[1]),
            "Metadata provider did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_associated_party(self):
        self.initialize_resource()
        self.assertEqual(
            0, len(self.empty_eml.resource.associated_party),
            "Associated Party from nowhere"
        )
        self.empty_eml.add_associated_party(
            ResponsibleParty(
                individual_name=IndividualName(first_name="Joe", last_name="Doe")
            ), Role.USER
        )
        self.empty_eml.add_associated_party(
            ResponsibleParty(
                individual_name=IndividualName(first_name="John", last_name="Doe")
            ), Role.ORIGINATOR
        )
        self.assertEqual(
            "Doe, J.", str(self.empty_eml.resource.associated_party[0][0]),
            "Associated Party did not set"
        )
        self.assertEqual(
            "Doe, J.", str(self.empty_eml.resource.associated_party[1][0]),
            "Associated Party did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_pub_date(self):
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.publication_date, "Publication date from nowhere")
        date = datetime.date.today()
        self.empty_eml.set_pub_date(date)
        self.assertEqual(
            date, self.empty_eml.resource.publication_date,
            "Publication date did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_language(self):
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.language, "Language from nowhere")
        self.empty_eml.set_language(Language.ENG)
        self.assertEqual(
            Language.ENG, self.empty_eml.resource.language,
            "Language did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_series(self):
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.series, "Series from nowhere")
        self.empty_eml.set_series("Volume 20")
        self.assertEqual(
            "Volume 20", self.empty_eml.resource.series,
            "Series did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_abstract(self):
        abstract = ("Lorem ipsum dolor sit amet, consectetur "
                    "adipiscing elit. Etiam vitae diam nec eros "
                    "interdum posuere. Class aptent taciti sociosqu "
                    "ad litora torquent per conubia nostra, "
                    "per inceptos himenaeos. Etiam vel dui "
                    "dignissim, aliquet felis ut, varius ligula. "
                    "Ut condimentum sapien in felis posuere, eu "
                    "aliquam enim lobortis. Vivamus sodales tortor "
                    "at augue gravida feugiat. Aenean faucibus "
                    "felis id aliquet euismod egestas.")
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.abstract, "Abstract from nowhere")
        self.empty_eml.set_abstract(EMLTextType(paragraphs=[abstract]))
        self.assertEqual(
            abstract, self.empty_eml.resource.abstract.paragraphs[0],
            "Abstract did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_keyword_set(self):
        self.initialize_resource()
        self.assertEqual(0, len(self.empty_eml.resource.keyword_set), "Keyword set from nowhere")
        self.empty_eml.add_keyword_set(EMLKeywordSet(["biodiversity"]))
        self.assertEqual(
            1, len(self.empty_eml.resource.keyword_set),
            "Keyword set did not set"
        )
        self.empty_eml.add_keyword_set(EMLKeywordSet(["checklist"]))
        self.assertEqual(
            2, len(self.empty_eml.resource.keyword_set),
            "Keyword set did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_additional_info(self):
        self.initialize_resource()
        self.assertEqual(0, len(self.empty_eml.resource.additional_info), "Additional information from nowhere")
        self.empty_eml.add_additional_info(EMLTextType())
        self.assertEqual(
            1, len(self.empty_eml.resource.additional_info),
            "Additional info did not set"
        )
        self.empty_eml.add_additional_info(EMLTextType())
        self.assertEqual(
            2, len(self.empty_eml.resource.additional_info),
            "Additional info did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_intellectual_rights(self):
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.intellectual_rights, "Intellectual rights from nowhere")
        self.empty_eml.set_intellectual_rights(EMLTextType(paragraphs=["CC-by"]))
        self.assertEqual(
            "CC-by", self.empty_eml.resource.intellectual_rights.paragraphs[0],
            "Intellectual rights did not set"
        )
        self.empty_eml.set_intellectual_rights(EMLTextType(paragraphs=["by-nc-nd"]))
        self.assertEqual(
            "by-nc-nd", self.empty_eml.resource.intellectual_rights.paragraphs[0],
            "Intellectual rights did not overwrite"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_licensed(self):
        self.initialize_resource()
        self.assertEqual(0, len(self.empty_eml.resource.licensed), "Licensed from nowhere")
        self.empty_eml.add_licensed(EMLLicense(name="Apache License 1.0"))
        self.assertEqual(
            1, len(self.empty_eml.resource.licensed),
            "Licensed did not set"
        )
        self.empty_eml.add_licensed(EMLLicense(name="Apache License 2.0"))
        self.assertEqual(
            2, len(self.empty_eml.resource.licensed),
            "Licensed did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_distribution(self):
        self.initialize_resource()
        self.assertEqual(0, len(self.empty_eml.resource.distribution), "Distribution from nowhere")
        self.empty_eml.add_distribution(EMLDistribution(online=EMLOnline(url="http://data.org/getdata?id=98332")))
        self.assertEqual(
            1, len(self.empty_eml.resource.distribution),
            "Distribution did not set"
        )
        self.validate(self.empty_eml.to_xml())
        self.empty_eml.add_distribution(EMLDistribution(offline=EMLOffline("Tape, 3.5 inch Floppy Disk, hardcopy")))
        self.assertEqual(
            2, len(self.empty_eml.resource.distribution),
            "Distribution did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_coverage(self):
        self.initialize_resource()
        self.assertIsNone(self.empty_eml.resource.coverage, "Coverage from nowhere")
        self.empty_eml.set_coverage(EMLCoverage(
            temporal=TemporalCoverage(single_datetime=[dt.date.today()])
        ))
        self.assertIsNotNone(
            self.empty_eml.resource.coverage,
            "Coverage did not set"
        )
        self.validate(self.empty_eml.to_xml())

    def test_set_annotation(self):
        self.initialize_resource(_id="1")
        self.assertEqual(0, len(self.empty_eml.resource.annotation), "Annotation from nowhere")
        self.empty_eml.add_annotation(SemanticAnnotation(
            ("http://example.org/height", "height"),
            ("http://example.org/tall", "tall"))
        )
        self.assertEqual(
            1, len(self.empty_eml.resource.annotation),
            "Annotation did not set"
        )
        self.empty_eml.add_annotation(SemanticAnnotation(
            ("http://example.org/widht", "width"),
            ("http://example.org/thick", "thick"))
        )
        self.assertEqual(
            2, len(self.empty_eml.resource.annotation),
            "Annotation did not set"
        )
        self.validate(self.empty_eml.to_xml())


if __name__ == '__main__':
    unittest.main()



from dwca.terms import Field


class MaterialSampleID(Field):
    """
    An identifier for the MaterialSample (as opposed to a particular digital record of the MaterialSample).
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/materialSampleID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

PATH = os.path.abspath(os.path.dirname(__file__))

sys.path.append(os.path.abspath(os.path.join(PATH, os.pardir, os.pardir, "src")))

project = 'Python Darwin Core Archive'
copyright = '2024, IEB-BIODATA'
author = 'Juan Saez Hidalgo'
release = '0.4.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.inheritance_diagram',
    'numpydoc',
]

autosummary_generate = True
numpydoc_show_class_members = True

templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'

html_static_path = ['_static',]
html_js_files = [
    '_js/pypi-icon.js',
]

html_title = "Python Darwin Core Archive"

html_sidebars = {
    "index": [],
    "install": [],
    "usage": [],
}

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/IEB-BIODATA/pydwca",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/pydwca/",
            "icon": "fa-custom fa-pypi",
        },
    ],
    "show_toc_level": 3
}

html_favicon = "favicon.png"



from __future__ import annotations

from typing import List, Dict, Type

from lxml import etree as et

from dwca.classes import DataFile, DataFileType
from dwca.terms import Field, OutsideTerm


class OutsideClass(DataFile):
    """
    Classes defined outside the Darwin Core specifications.

    Parameters
    ----------
    _id : int
        Unique identifier for the core entity.
    uri : str
        URI of the term.
    files : str
        File location, in the archive, this is inside the `zip` file.
    data_file_type: DataFileType
        The Data File Type in the Darwin Core Archive.
    fields : List[Field]
        A list of the Field (columns) in the Core data entity.
    encoding : str, optional
        Encoding of the file location (`files` parameter), default is "utf-8".
    lines_terminated_by : str, optional
        Delimiter of lines on the file, default `"\\\\n"`.
    fields_terminated_by : str, optional
        Delimiter of the file (cells) on the file, default `","`.
    fields_enclosed_by : str, optional
        Specifies the character used to enclose (mark the start and end of) each field, default empty `""`.
    ignore_header_lines : int, optional
        Ignore headers at the start of document, can be one line or a list of them, default 0 (first line).
    """
    __field_class__ = DataFile.__field_class__
    def __init__(
            self, _id: int, uri: str,
            files: str,
            fields: List[Field],
            data_file_type: DataFileType = DataFileType.CORE,
            encoding: str = "utf-8",
            lines_terminated_by: str = "\n",
            fields_terminated_by: str = ",",
            fields_enclosed_by: str = "",
            ignore_header_lines: int = 0,
    ) -> None:
        super().__init__(
            _id, files, fields, data_file_type, encoding,
            lines_terminated_by, fields_terminated_by,
            fields_enclosed_by, ignore_header_lines,
        )
        self.URI = uri
        return

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> OutsideClass | None:
        """
        Parse an `lxml.etree.Element` into an OutsideClass object.

        Parameters
        ----------
        element : `lxml.etree.Element`
            An XML `Element`.
        nmap : Dict
            Dictionary of prefix:uri.

        Returns
        -------
        OutsideClass
            An instance of OutsideClass.
        """
        if element is None:
            return None
        kwargs = cls.parse_kwargs(element, nmap)
        kwargs["uri"] = element.get("rowType")
        outside = OutsideClass(**kwargs)
        outside.__namespace__ = nmap
        return outside

    @classmethod
    def get_term_class(cls, element: et.Element) -> Type[Field]:
        """
        Extract the Python ``class`` term from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element instance.

        Returns
        -------
        Type[Field]
            The Python ``class`` representing the term name.
        """
        for field_class in cls.__field_class__:
            if element.get("term") == field_class.URI:
                return field_class
        return OutsideTerm



from xml_common.utils.iteratate import iterate_with_bar, OptionalTqdm
from xml_common.utils.enum import CamelCaseEnum
from xml_common.utils.language import Language
from xml_common.utils.read_file import read_string
from xml_common.utils.gpolygon import GRing, GPolygon
from xml_common.utils.length_unit import LengthUnit
from xml_common.utils.type_functions import format_to_type, format_union, format_datetime
from xml_common.utils.type_functions import unformat_type
from xml_common.utils.type_functions import type_to_pl
from xml_common.utils.type_functions import type_to_sql, format_to_sql
from xml_common.utils.establishment_means import EstablishmentMeans



from __future__ import annotations

from typing import Dict, List

from lxml import etree as et

from eml.resources import Resource
from eml.types import ResponsibleParty, EMLTextType, Maintenance, Methods, ResearchProject


class EMLDataset(Resource):
    """
    Dataset represents the base type for the dataset element on an EML document.

    Parameters
    ----------
    purpose : EMLTextType, optional
        A synopsis of the purpose of this dataset.
    introduction : EMLTextType, optional
        An overview of the background and context for the dataset.
    getting_started : EMLTextType, optional
        A high level overview of interpretation, structure, and content of the dataset.
    acknowledgement : EMLTextType, optional
        Text that acknowledges funders and other key contributors.
    maintenance : Maintenance, optional
        A description of the maintenance of this data resource.
    contact : List[ResponsibleParty]
        The contact for this dataset.
    publisher : ResponsibleParty, optional
        The publisher of this data set.
    pub_place : str, optional
        The location that the resource was published.
    methods : Methods, optional
        The methods field documents scientific methods used in the collection of this dataset.
    project : ResearchProject, optional
        The project in which this dataset was collected.

    Other Parameters
    ----------------
    **kwargs : :class:`eml.resources.resource.Resource` parameters.
        The parameters of every type of Resource.
    """
    PRINCIPAL_TAG = "dataset"

    def __init__(
            self,
            purpose: EMLTextType = None,
            introduction: EMLTextType = None,
            getting_started: EMLTextType = None,
            acknowledgements: EMLTextType = None,
            maintenance: Maintenance = None,
            contact: List[ResponsibleParty] = None,
            publisher: ResponsibleParty = None,
            pub_place: str = None,
            methods: Methods = None,
            project: ResearchProject = None,
            **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.__contact__ = list()
        if self.referencing:
            return
        if contact is None or len(contact) == 0:
            raise ValueError("No contact provided.")
        self.__contact__.extend(contact)
        self.__purpose__ = purpose
        self.__intro__ = introduction
        self.__get_started__ = getting_started
        self.__acknowledge__ = acknowledgements
        self.__maint__ = maintenance
        self.__publisher__ = publisher
        self.__pub_place__ = pub_place
        self.__methods__ = methods
        self.__project__ = project
        return

    @property
    def contacts(self) -> List[ResponsibleParty]:
        """List[ResponsibleParty]: The contact for this dataset."""
        return self.__contact__

    @property
    def purpose(self) -> EMLTextType:
        """EMLTextType: A synopsis of the purpose of this dataset."""
        return self.__purpose__

    @property
    def introduction(self) -> EMLTextType:
        """EMLTextType: An overview of the background and context for the dataset."""
        return self.__intro__

    @property
    def getting_started(self) -> EMLTextType:
        """EMLTextType: A high level overview of interpretation, structure, and content of the dataset."""
        return self.__get_started__

    @property
    def acknowledgements(self) -> EMLTextType:
        """EMLTextType: Text that acknowledges funders and other key contributors."""
        return self.__acknowledge__

    @property
    def maintenance(self) -> Maintenance:
        """Maintenance: A description of the maintenance of this data resource."""
        return self.__maint__

    @property
    def publisher(self) -> ResponsibleParty:
        """ResponsibleParty: The publisher of this data set."""
        return self.__publisher__

    @property
    def pub_place(self) -> str:
        """str: The location that the resource was published."""
        return self.__pub_place__

    @property
    def methods(self) -> Methods:
        """Methods: The methods field documents scientific methods used in the collection of this dataset."""
        return self.__methods__

    @property
    def project(self) -> ResearchProject:
        """ResearchProject: The project in which this dataset was collected."""
        return self.__project__

    # TODO: Implement other attributes (https://eml.ecoinformatics.org/schema/eml_xsd#eml_dataset)

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLDataset:
        """
        Generate an EML Dataset referencing another EML Dataset.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references another dataset.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLDataset
            Object parsed that reference another dataset.
        """
        kwargs = super().parse_kwargs(element, nmap)
        return EMLDataset(**kwargs)

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLDataset:
        """
        Generate an EML Dataset that do not reference another dataset.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLDataset
            Object parsed.
        """
        kwargs = super().parse_kwargs(element, nmap)
        purpose_elem = element.find("purpose", nmap)
        purpose = None if purpose_elem is None else EMLTextType.parse(purpose_elem, nmap)
        intro_elem = element.find("introduction", nmap)
        introduction = None if intro_elem is None else EMLTextType.parse(intro_elem, nmap)
        get_started_elem = element.find("gettingStarted", nmap)
        get_started = None if get_started_elem is None else EMLTextType.parse(get_started_elem, nmap)
        acknowledge_elem = element.find("acknowledgements", nmap)
        acknowledgements = None if acknowledge_elem is None else EMLTextType.parse(acknowledge_elem, nmap)
        maint_elem = element.find("maintenance", nmap)
        maint = None if maint_elem is None else Maintenance.parse(maint_elem, nmap)
        contact = list()
        for contact_elem in element.findall("contact", nmap):
            contact.append(ResponsibleParty.parse(contact_elem, nmap))
        publisher_elem = element.find("publisher", nmap)
        publisher = None if publisher_elem is None else ResponsibleParty.parse(publisher_elem, nmap)
        pub_place_elem = element.find("pubPlace", nmap)
        pub_place = None if pub_place_elem is None else pub_place_elem.text
        methods_elem = element.find("methods", nmap)
        methods = None if methods_elem is None else Methods.parse(methods_elem, nmap)
        project_elem = element.find("project", nmap)
        project = None if project_elem is None else ResearchProject.parse(project_elem, nmap)
        return EMLDataset(
            purpose=purpose,
            introduction=introduction,
            getting_started=get_started,
            acknowledgements=acknowledgements,
            maintenance=maint,
            contact=contact,
            publisher=publisher,
            pub_place=pub_place,
            methods=methods,
            project=project,
            **kwargs
        )

    def to_element(self) -> et.Element:
        """
        Generates an XML `Element` from this instance

        Returns
        -------
        `lxml.etree.Element`
            XML `Element` from this instance
        """
        dataset = super().to_element()
        if self.referencing:
            return dataset
        for contact in self.contacts:
            contact.set_tag("contact")
            dataset.append(contact.to_element())
        if self.purpose is not None:
            self.purpose.set_tag("purpose")
            dataset.append(self.purpose.to_element())
        if self.introduction is not None:
            self.introduction.set_tag("introduction")
            dataset.append(self.introduction.to_element())
        if self.getting_started is not None:
            self.getting_started.set_tag("gettingStarted")
            dataset.append(self.getting_started.to_element())
        if self.acknowledgements is not None:
            self.acknowledgements.set_tag("acknowledgements")
            dataset.append(self.acknowledgements.to_element())
        if self.maintenance is not None:
            dataset.append(self.maintenance.to_element())
        if self.publisher is not None:
            self.publisher.set_tag("publisher")
            dataset.append(self.publisher.to_element())
        if self.pub_place is not None:
            pub_place_elem = self.object_to_element("pubPlace")
            pub_place_elem.text = self.pub_place
            dataset.append(pub_place_elem)
        if self.methods is not None:
            dataset.append(self.methods.to_element())
        if self.project is not None:
            self.project.set_tag("project")
            dataset.append(self.project.to_element())
        return dataset



import unittest

from xml_common.utils import EstablishmentMeans


class TestEstablishmentMeans(unittest.TestCase):
    def test_camel_case(self):
        native = EstablishmentMeans(1)
        self.assertEqual("native", native.to_camel_case(), "Error on one word establishment")
        assisted = EstablishmentMeans(4)
        self.assertEqual(
            "introducedAssistedColonisation",
            assisted.to_camel_case(),
            "Error on multiple word establishment"
        )

    def test_iri(self):
        native = EstablishmentMeans.NATIVE
        self.assertEqual(
            "http://rs.tdwg.org/dwcem/values/e001",
            native.get_iri(),
            "Error on uri generation"
        )

    def test_get_establishment(self):
        self.assertEqual(
            EstablishmentMeans.NATIVE,
            EstablishmentMeans.get_enum("native"),
            "Error on one word establishment")
        self.assertEqual(
            EstablishmentMeans.INTRODUCED_ASSISTED_COLONISATION,
            EstablishmentMeans.get_enum("introducedAssistedColonisation"),
            "Error on multiple word establishment"
        )


if __name__ == '__main__':
    unittest.main()



from dwca.terms import Field


class GeologicalContextID(Field):
    """
    An identifier for the set of information associated with a GeologicalContext.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/geologicalContextID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestEonOrLowestEonothem(Field):
    """
    The earliest possible geochronologic eon or lowest chrono-stratigraphic eonothem.

    The full name of the earliest possible geochronologic eon
    or lowest chrono-stratigraphic eonothem or the informal
    name ("Precambrian") attributable to the stratigraphic
    horizon from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/earliestEonOrLowestEonothem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LatestEonOrHighestEonothem(Field):
    """
    The latest possible geochronologic eon or highest chrono-stratigraphic eonothem.

    The full name of the latest possible geochronologic eon
    or highest chrono-stratigraphic eonothem or the informal
    name ("Precambrian") attributable to the stratigraphic
    horizon from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/latestEonOrHighestEonothem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestEraOrLowestErathem(Field):
    """
    The earliest possible geochronologic era or lowest chronostratigraphic erathem.

    The full name of the earliest possible geochronologic era
    or lowest chronostratigraphic erathem attributable to the
    stratigraphic horizon from which the MaterialEntity was
    collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/earliestEraOrLowestErathem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LatestEraOrHighestErathem(Field):
    """
    The latest possible geochronologic era or highest chronostratigraphic erathem.

    The full name of the latest possible geochronologic era
    or highest chronostratigraphic erathem attributable to
    the stratigraphic horizon from which the MaterialEntity
    was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/latestEraOrHighestErathem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestPeriodOrLowestSystem(Field):
    """
    The earliest possible geochronologic period or lowest chronostratigraphic system.

    The full name of the earliest possible geochronologic period or
    lowest chronostratigraphic system attributable to the
    stratigraphic horizon from which the MaterialEntity was
    collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/earliestPeriodOrLowestSystem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LatestPeriodOrHighestSystem(Field):
    """
    The latest possible geochronologic period or highest chronostratigraphic system.

    The full name of the latest possible geochronologic period or
    highest chronostratigraphic system attributable to the
    stratigraphic horizon from which the MaterialEntity
    was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/latestPeriodOrHighestSystem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestEpochOrLowestSeries(Field):
    """
    The earliest possible geochronologic epoch or lowest chronostratigraphic series.

    The full name of the earliest possible geochronologic epoch
    or lowest chronostratigraphic series attributable to the
    stratigraphic horizon from which the MaterialEntity was
    collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/earliestEpochOrLowestSeries"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LatestEpochOrHighestSeries(Field):
    """
    The latest possible geochronologic epoch or highest chronostratigraphic series.

    The full name of the latest possible geochronologic epoch or
    highest chronostratigraphic series attributable to the
    stratigraphic horizon from which the MaterialEntity was
    collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/latestEpochOrHighestSeries"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestAgeOrLowestStage(Field):
    """
    The earliest possible geochronologic age or lowest chronostratigraphic stage.

    The full name of the earliest possible geochronologic age
    or lowest chronostratigraphic stage attributable to the
    stratigraphic horizon from which the MaterialEntity was
    collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/earliestAgeOrLowestStage"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LatestAgeOrHighestStage(Field):
    """
    The latest possible geochronologic age or highest chronostratigraphic stage.

    The full name of the latest possible geochronologic age or
    highest chronostratigraphic stage attributable to the
    stratigraphic horizon from which the MaterialEntity was
    collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/latestAgeOrHighestStage"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LowestBiostratigraphicZone(Field):
    """
    The lowest possible geological biostratigraphic zone.

    The full name of the lowest possible geological
    biostratigraphic zone of the stratigraphic horizon
    from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/lowestBiostratigraphicZone"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class HighestBiostratigraphicZone(Field):
    """
    The highest possible geological biostratigraphic zone.

    The full name of the highest possible geological
    biostratigraphic zone of the stratigraphic horizon
    from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/highestBiostratigraphicZone"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LithostratigraphicTerms(Field):
    """
    The combination of all litho-stratigraphic names for the rock.

    The combination of all litho-stratigraphic names for the rock
    from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/lithostratigraphicTerms"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LithostratigraphicGroup(Field):
    """
    The full name of the lithostratigraphic group from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/group"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LithostratigraphicFormation(Field):
    """
    The full name of the lithostratigraphic formation from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/formation"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LithostratigraphicMember(Field):
    """
    The full name of the lithostratigraphic member from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/member"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LithostratigraphicBed(Field):
    """
    The full name of the lithostratigraphic bed from which the MaterialEntity was collected.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/bed"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import Dict, Union

from lxml import etree as et


class XMLObject(ABC):
    """
    Abstract class for XML Object, to be parsed on the read of the Darwin Core Archive file.
    """
    PRINCIPAL_TAG = "placeholder"
    """str : The principal tag of the XML document."""
    NAMESPACE_TAG = None
    """str : If the principal tag has a namespace attribute, this must be used."""
    NAMESPACES = {
        "xml": "http://www.w3.org/XML/1998/namespace"
    }
    """Dict[str, str] : Namespaces to use on the particular XML portion that represent the object."""

    def __init__(self) -> None:
        self.__namespace__ = dict()
        for prefix, uri in self.NAMESPACES.items():
            self.__namespace__[prefix] = uri
        return

    @classmethod
    def from_xml(cls, file: str, encoding: str = "utf-8") -> XMLObject:
        """
        Generates an XML Object from an XML file.

        Parameters
        ----------
        file : str
            Path to XML file.
        encoding : str, default "utf-8"
            Encoding of the XML file.

        Returns
        -------
        XMLObject
            Parsed XML object from XML file.
        """
        with open(file, "r", encoding=encoding) as file:
            content = file.read()
        return cls.from_string(content)

    @classmethod
    def from_string(cls, text: str) -> XMLObject:
        """
        Generates XML Object from a string of an XML file.

        Parameters
        ----------
        text : str
            XML File as a string.

        Returns
        -------
        XMLObject
            Object parsed from the string.
        """
        cleaned_text = re.sub(r'<\?xml\s+version="1.0"\s+encoding="[^"]*"\?>', '', text, count=1)
        if text != cleaned_text and cleaned_text.endswith("</xml>"):
            cleaned_text = cleaned_text[:-6]
        root = et.fromstring(cleaned_text)
        nmap = dict()
        for prefix, uri in cls.NAMESPACES.items():
            nmap[prefix] = uri
        for prefix, uri in root.nsmap.items():
            nmap[prefix] = uri
        cls.check_principal_tag(root.tag, nmap)
        xml_object = cls.parse(root, nmap)
        xml_object.__namespace__ = nmap
        return xml_object

    @classmethod
    @abstractmethod
    def parse(cls, element: et.Element, nmap: Dict) -> XMLObject | None:
        """
        Parses a lxml element into an XMLObject.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        XMLObject
            Object parsed.
        """
        pass

    def to_xml(self) -> str:
        """
        Generates text of an XML file.

        Returns
        -------
        str
            Text of an XML File from the object.
        """
        return et.tostring(self.to_element(), pretty_print=True).decode()

    @abstractmethod
    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the object.

        Returns
        -------
        lxml.tree.Element
            Object in the Element format.
        """
        return et.Element(self.get_principal_tag(), nsmap=self.__namespace__)

    def object_to_element(self, tag: str, prefix: Union[str, None] = None) -> et.Element:
        """
        Generates an element using tag, adding namespace tag.

        Parameters
        ----------
        tag : str
            Tag for XML Element.
        prefix : str, optional
            Prefix of namespace, default None.

        Returns
        -------
        lxml.tree.Element
            Generated element.
        """
        if prefix in self.__namespace__:
            return et.Element(f"{{{self.__namespace__[prefix]}}}{tag}")
        else:
            return et.Element(tag)

    def get_principal_tag(self) -> str:
        """
        Returns the principal tag with namespaces if it is present.

        Returns
        -------
        str
            Principal tag of the XML file.
        """
        if self.NAMESPACE_TAG is not None:
            return f"{{{self.__namespace__[self.NAMESPACE_TAG]}}}{self.PRINCIPAL_TAG}"
        if None in self.__namespace__:
            return f"{{{self.__namespace__[None]}}}{self.PRINCIPAL_TAG}"
        return self.PRINCIPAL_TAG

    @classmethod
    def check_principal_tag(cls, tag: str, nmap: Dict) -> None:
        """
        Checks if the tag is the Principal tag of the object.

        Parameters
        ----------
        tag : str
            Tag obtained from outside source.
        nmap : Dict
            The namespace to check the format of the Principal tag.

        Raises
        ------
        AssertionError
            If the tag is not a representation of the principal tag of the object.
        """
        if cls.NAMESPACE_TAG is not None:
            expected = f"{{{nmap[cls.NAMESPACE_TAG]}}}{cls.PRINCIPAL_TAG}"
        elif None in nmap:
            expected = f"{{{nmap[None]}}}{cls.PRINCIPAL_TAG}"
        else:
            expected = f"{cls.PRINCIPAL_TAG}"
        assert tag == expected, f"{tag} is not {expected}"

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"<XMLObject tag={self.PRINCIPAL_TAG}>"



import datetime as dt
from typing import Union

from datetime_interval import Interval

from dwca.terms import Field


class ResourceRelationshipID(Field):
    """
    An identifier for an instance of relationship between one resource (the subject) and another (the object).
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/resourceRelationshipID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ResourceID(Field):
    """
    An identifier for the resource that is the subject of the relationship.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/resourceID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RelationshipOfResourceID(Field):
    """
    An identifier for the relationship type.

    An identifier for the relationship type
    (predicate) that connects the subject
    identified by Resource ID to its object
    identified by Related Resource ID.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/relationshipOfResourceID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class RelatedResourceID(Field):
    """
    An identifier for a related resource.

    The object, rather than the subject of the relationship.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/relatedResourceID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RelationshipOfResource(Field):
    """
    The relationship of the subject to the object.

    The subject identified by ``Resource ID``. The object
    identified by ``Related Resource ID``.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/relationshipOfResource"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class RelationshipAccordingTo(Field):
    """
    The source establishing the relationship between the two resources.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/relationshipAccordingTo"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RelationshipEstablishedDate(Field):
    """
    The date-time on which the relationship between the two resources was established.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/relationshipEstablishedDate"
    TYPE = Union[dt.datetime, Interval]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RelationshipRemarks(Field):
    """
    Comments or notes about the relationship between the two resources.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/relationshipRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



from __future__ import annotations

from typing import Dict, List, Union

from lxml import etree as et

from xml_common import XMLObject


class EMLMetadata(XMLObject):
    """
    This element contains the additional metadata that is to be included in the document.

    Parameters
    ----------
    xml: lxml.etree.Element or str, optional
        An XML element instance or string representing a XML well define file.
    """
    PRINCIPAL_TAG = "metadata"

    def __init__(
            self,
            xml: Union[et.Element, str] = None
    ) -> None:
        super().__init__()
        if isinstance(xml, str):
            self.__content__ = et.fromstring(xml)
        else:
            self.__content__ = xml
        return

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLMetadata | None:
        """
        Generate a EMLMetadata from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLMetadata
            An EMLMetadata object.
        """
        if element is None:
            return None
        metadata = EMLMetadata(xml=element)
        metadata.__namespace__ = nmap
        return metadata

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance from this EMLMetadata object.

        Returns
        -------
        lxml.etree.Element
            An XML element instance.
        """
        return self.__content__


class EMLAdditionalMetadata(XMLObject):
    """
    A flexible field for including any other relevant metadata that pertains to the resource being described.

    Parameters
    ----------
    metadata: EMLMetadata
        This element contains the additional metadata that is to be included in the document.
    _id: str, optional
        Unique identifier for the additional metadata.
    describes: List[str], optional
        A pointer to the id attribute for the sub-portion of the resource that is described by this additional metadata.
    """
    PRINCIPAL_TAG = "additionalMetadata"

    def __init__(self, metadata: EMLMetadata, _id: str = None, describes: List[str] = None) -> None:
        super().__init__()
        self.__metadata__ = metadata
        self.__id__ = _id
        self.__describes__ = list()
        if describes is not None:
            self.__describes__.extend(describes)
        return

    @property
    def id(self) -> str:
        """str: Unique identifier for the additional metadata."""
        return self.__id__


    @property
    def metadata(self) -> EMLMetadata:
        """EMLMetadata: This element contains the additional metadata that is to be included in the document."""
        return self.__metadata__

    @property
    def describes(self) -> List[str]:
        """List[str]: A pointer to the id attribute for the sub-portion of the resource."""
        return self.__describes__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLAdditionalMetadata | None:
        """
        Generate an EMLAdditionalMetadata object from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLAdditionalMetadata
            An EMLAdditionalMetadata object.
        """
        if element is None:
            return None
        describes = list()
        for describe_elem in element.findall("describes", nmap):
            describes.append(describe_elem.text)
        metadata = EMLAdditionalMetadata(
            metadata=EMLMetadata.parse(element.find("metadata", nmap), nmap),
            _id=element.get("id", None),
            describes=describes,
        )
        metadata.__namespace__ = nmap
        return metadata

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance from the AdditionalMetadata object.

        Returns
        -------
        lxml.etree.Element
            An XML element instance.
        """
        root = et.Element("additionalMetadata")
        root.append(self.metadata.to_element())
        if self.id is not None:
            root.set("id", self.id)
        for describe in self.describes:
            describe_elem = self.object_to_element("describes")
            describe_elem.text = describe
            root.append(describe_elem)
        return root



import unittest

from lxml import etree as et

from eml.resources.coverage import TaxonomicCoverage
from test_xml.test_xml import TestXML


class MyTestCase(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
        '{http://www.w3.org/XML/1998/namespace}lang': 'eng',
    }

    def test_parse(self):
        general_taxonomic_coverage = ("All vascular plants were identified "
                                      "to family or species, mosses and "
                                      "lichens were identified as moss or "
                                      "lichen.")
        text_xml = f"""
<taxonomicCoverage id="" scope="document" system="">
    <taxonomicSystem>
        <classificationSystem>
            <classificationSystemCitation>
                <references>class-system</references>
            </classificationSystemCitation>
        </classificationSystem>
        <identifierName>
            <organizationName>Identifier Organization</organizationName>
        </identifierName>
        <taxonomicProcedures>specimen processing</taxonomicProcedures>
    </taxonomicSystem>
    <generalTaxonomicCoverage>{general_taxonomic_coverage}</generalTaxonomicCoverage>
    <taxonomicClassification>
        <taxonRankName>Kingdom</taxonRankName>
        <taxonRankValue>Viridiplantae</taxonRankValue>
        <commonName>Green Plants</commonName>
        <taxonId provider="https://www.ncbi.nlm.nih.gov/taxonomy">33090</taxonId>
    </taxonomicClassification>
</taxonomicCoverage>
        """
        coverage = TaxonomicCoverage.from_string(text_xml)
        self.assertEqual(
            1, len(coverage.taxonomic_system.classification_system),
            "Error on parsing classification system"
        )
        self.assertEqual(
            "class-system", coverage.taxonomic_system.classification_system[0].citation.id,
            "Error on parse classification system citation"
        )
        self.assertIsNone(
            coverage.taxonomic_system.classification_system[0].modifications,
            "Error on parse classification system modifications"
        )
        self.assertEqual(
            0, len(coverage.taxonomic_system.identification_reference),
            "Identification reference from nowhere"
        )
        self.assertEqual(
            1, len(coverage.taxonomic_system.identifier_name),
            "Error on parsing identifier name"
        )
        self.assertEqual(
            "Identifier Organization", coverage.taxonomic_system.identifier_name[0].organization_name,
            "Error on parsing identifier name (organization name)"
        )
        self.assertEqual(
            "specimen processing", coverage.taxonomic_system.procedures,
            "Error on parsing taxonomic procedures"
        )
        self.assertEqual(
            general_taxonomic_coverage,
            coverage.general_coverage,
            "Error on parsing general taxonomic coverage"
        )
        self.assertEqual(
            1, len(coverage.classification),
            "Error on parsing classification system"
        )
        self.assertEqual("Kingdom", coverage.classification[0].rank_name, "Error on parsing taxonomic rank name")
        self.assertEqual("Viridiplantae", coverage.classification[0].rank_value, "Error on parsing taxonomic rank value")
        self.assertEqual(1, len(coverage.classification[0].common_name), "Error on parsing xml_common name")
        self.assertEqual("Green Plants", coverage.classification[0].common_name[0], "Error on parsing xml_common name")
        self.assertEqual(1, len(coverage.classification[0].taxon_id), "Error on parsing taxon id")
        self.assertEqual(
            33090,
            coverage.classification[0].taxon_id[0],
            "Error on parsing taxon id"
        )
        self.assertEqual(
            "https://www.ncbi.nlm.nih.gov/taxonomy",
            coverage.classification[0].taxon_id[0].provider,
            "Error on parsing taxon id"
        )
        self.assertEqual(
            0, len(coverage.classification[0].classification),
            "Taxonomic classification from nowhere"
        )
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_parse_taxonomic_system(self):
        text_xml = """
<taxonomicSystem>
    <classificationSystem>
        <classificationSystemCitation>
            <references>class-system</references>
        </classificationSystemCitation>
        <classificationSystemModifications>Description of any modifications or exceptions.</classificationSystemModifications>
    </classificationSystem>
    <identificationReference>
        <references>knb.45.3</references>
    </identificationReference>
    <identifierName>
        <organizationName>Identifier Organization</organizationName>
    </identifierName>
    <taxonomicProcedures>specimen processing</taxonomicProcedures>
    <taxonomicCompleteness>materials sent to experts</taxonomicCompleteness>
    <vouchers>
        <specimen>herbarium specimens</specimen>
        <repository>
            <originator>
                <organizationName>Repository Organization</organizationName>
            </originator>
        </repository>
    </vouchers>
</taxonomicSystem>
        """
        taxonomic_system = TaxonomicCoverage.TaxonomicSystem.from_string(text_xml)
        self.assertEqual(
            1, len(taxonomic_system.classification_system),
            "Error on parsing classification system"
        )
        self.assertEqual(
            "class-system", taxonomic_system.classification_system[0].citation.id,
            "Error on parse classification system citation"
        )
        self.assertEqual(
            "Description of any modifications or exceptions.",
            taxonomic_system.classification_system[0].modifications,
            "Error on parse classification system modifications"
        )
        self.assertEqual(
            1, len(taxonomic_system.identification_reference),
            "Error on parsing identification reference"
        )
        self.assertEqual(
            "knb.45.3", taxonomic_system.identification_reference[0].id,
            "Error on parse identification reference"
        )
        self.assertEqual(
            1, len(taxonomic_system.identifier_name),
            "Error on parsing identifier name"
        )
        self.assertEqual(
            "Identifier Organization", taxonomic_system.identifier_name[0].organization_name,
            "Error on parsing identifier name (organization name)"
        )
        self.assertEqual(
            "specimen processing", taxonomic_system.procedures,
            "Error on parsing taxonomic procedures"
        )
        self.assertEqual(
            "materials sent to experts", taxonomic_system.completeness,
            "Error on parsing taxonomic completeness"
        )
        self.assertEqual(
            1, len(taxonomic_system.vouchers),
            "Error on parsing vouchers"
        )
        self.assertEqual(
            "herbarium specimens", taxonomic_system.vouchers[0].specimen,
            "Error on parsing vouchers specimen"
        )
        self.assertEqual(
            1, len(taxonomic_system.vouchers[0].repository),
            "Error on parsing vouchers repository"
        )
        self.assertEqual(
            "Repository Organization", taxonomic_system.vouchers[0].repository[0].organization_name,
            "Error on parsing vouchers repository"
        )
        self.assertEqualTree(et.fromstring(text_xml), taxonomic_system.to_element(), "Error on to element")

    def test_parse_taxonomic_system_invalid(self):
        self.assertRaises(
            ValueError, TaxonomicCoverage.TaxonomicSystem.from_string,
            """
<taxonomicSystem>
    <identificationReference>
        <references>knb.45.3</references>
    </identificationReference>
    <identifierName>
        <organizationName>Identifier Organization</organizationName>
    </identifierName>
    <taxonomicProcedures>specimen processing</taxonomicProcedures>
    <taxonomicCompleteness>materials sent to experts</taxonomicCompleteness>
    <vouchers>
        <specimen>herbarium specimens</specimen>
        <repository>
            <originator>
                <organizationName>Repository Organization</organizationName>
            </originator>
        </repository>
    </vouchers>
</taxonomicSystem>
            """
        )
        self.assertRaises(
            ValueError, TaxonomicCoverage.TaxonomicSystem.from_string,
            """
<taxonomicSystem>
    <classificationSystem>
        <classificationSystemCitation>
            <references>class-system</references>
        </classificationSystemCitation>
        <classificationSystemModifications>Description of any modifications or exceptions.</classificationSystemModifications>
    </classificationSystem>
    <identificationReference>
        <references>knb.45.3</references>
    </identificationReference>
    <taxonomicProcedures>specimen processing</taxonomicProcedures>
    <taxonomicCompleteness>materials sent to experts</taxonomicCompleteness>
    <vouchers>
        <specimen>herbarium specimens</specimen>
        <repository>
            <originator>
                <organizationName>Repository Organization</organizationName>
            </originator>
        </repository>
    </vouchers>
</taxonomicSystem>
            """
        )

    def test_parse_taxonomic_system_none(self):
        self.assertIsNone(TaxonomicCoverage.TaxonomicSystem.parse(None, {}), "Taxonomic System from nowhere")

    def test_parse_taxonomic_classification(self):
        text_xml = """
<taxonomicClassification id="">
    <taxonRankName>Kingdom</taxonRankName>
    <taxonRankValue>Viridiplantae</taxonRankValue>
    <commonName>Green Plants</commonName>
    <taxonId provider="https://www.ncbi.nlm.nih.gov/taxonomy">33090</taxonId>
    <taxonomicClassification>
        <taxonRankName>Phylum</taxonRankName>
        <taxonRankValue>Chlorophyta</taxonRankValue>
        <commonName>Green Algae</commonName>
        <taxonId provider="https://www.ncbi.nlm.nih.gov/taxonomy">3041</taxonId>
    </taxonomicClassification>
</taxonomicClassification>
        """
        taxonomic_classification = TaxonomicCoverage.TaxonomicClassification.from_string(text_xml)
        self.assertEqual("Kingdom", taxonomic_classification.rank_name, "Error on parsing taxonomic rank name")
        self.assertEqual("Viridiplantae", taxonomic_classification.rank_value, "Error on parsing taxonomic rank value")
        self.assertEqual(1, len(taxonomic_classification.common_name), "Error on parsing xml_common name")
        self.assertEqual("Green Plants", taxonomic_classification.common_name[0], "Error on parsing xml_common name")
        self.assertEqual(1, len(taxonomic_classification.taxon_id), "Error on parsing taxon id")
        self.assertEqual(
            33090,
            taxonomic_classification.taxon_id[0],
            "Error on parsing taxon id"
        )
        self.assertEqual(
            "https://www.ncbi.nlm.nih.gov/taxonomy",
            taxonomic_classification.taxon_id[0].provider,
            "Error on parsing taxon id"
        )
        self.assertEqual(
            1, len(taxonomic_classification.classification),
            "Error on parsing taxonomic classification"
        )
        self.assertEqual(
            "Phylum", taxonomic_classification.classification[0].rank_name,
            "Error on parsing taxonomic rank name"
        )
        self.assertEqual(
            "Chlorophyta",
            taxonomic_classification.classification[0].rank_value,
            "Error on parsing taxonomic rank value"
        )
        self.assertEqual(1, len(taxonomic_classification.classification[0].common_name), "Error on parsing xml_common name")
        self.assertEqual("Green Algae", taxonomic_classification.classification[0].common_name[0], "Error on parsing xml_common name")
        self.assertEqual(1, len(taxonomic_classification.classification[0].taxon_id), "Error on parsing taxon id")
        self.assertEqual(
            3041,
            taxonomic_classification.classification[0].taxon_id[0],
            "Error on parsing taxon id"
        )
        self.assertEqual(
            "https://www.ncbi.nlm.nih.gov/taxonomy",
            taxonomic_classification.classification[0].taxon_id[0].provider,
            "Error on parsing taxon id"
        )
        self.assertEqual(
            0, len(taxonomic_classification.classification[0].classification),
            "Taxonomic classification from nowhere"
        )
        self.assertEqualTree(et.fromstring(text_xml), taxonomic_classification.to_element(), "Error on to element")

    def test_parse_taxonomic_classification_none(self):
        self.assertIsNone(
            TaxonomicCoverage.TaxonomicClassification.parse(None, {}),
            "Taxonomic classification from nowhere"
        )

    def test_equal_taxon_id(self):
        example = TaxonomicCoverage.TaxonID(1, "http://gbif.org")
        self.assertEqual(1, example, "Not compare to int")
        same_example = TaxonomicCoverage.TaxonID(1, "http://gbif.org")
        self.assertEqual(example, same_example, "Same dato, not equals")
        other_example = TaxonomicCoverage.TaxonID(1, "http://gbif_alternative.org")
        self.assertNotEqual(example, other_example, "Equal different provider")
        self.assertNotEqual("1", example, "Error on equal implementation")

    def test_parse_invalid(self):
        general_taxonomic_coverage = ("All vascular plants were identified "
                                      "to family or species, mosses and "
                                      "lichens were identified as moss or "
                                      "lichen.")
        text_xml = f"""
<taxonomicCoverage id="" scope="document" system="">
    <taxonomicSystem>
        <classificationSystem>
            <classificationSystemCitation>
                <references>class-system</references>
            </classificationSystemCitation>
        </classificationSystem>
        <identifierName>
            <organizationName>Identifier Organization</organizationName>
        </identifierName>
        <taxonomicProcedures>specimen processing</taxonomicProcedures>
    </taxonomicSystem>
    <generalTaxonomicCoverage>{general_taxonomic_coverage}</generalTaxonomicCoverage>
</taxonomicCoverage>
        """
        self.assertRaises(ValueError, TaxonomicCoverage.from_string, text_xml)

    def test_parse_none(self):
        self.assertIsNone(TaxonomicCoverage.parse(None, {}), "Taxonomic Coverage from nowhere")

    def test_parse_minimum(self):
        text_xml = """
<taxonomicCoverage>
    <taxonomicClassification>
        <taxonRankName>Kingdom</taxonRankName>
        <taxonRankValue>Viridiplantae</taxonRankValue>
        <commonName>Green Plants</commonName>
        <taxonId provider="https://www.ncbi.nlm.nih.gov/taxonomy">33090</taxonId>
    </taxonomicClassification>
</taxonomicCoverage>
        """
        coverage = TaxonomicCoverage.from_string(text_xml)
        self.assertIsNone(coverage.id, "Identifier from nowhere")
        self.assertFalse(coverage.referencing, "Error on parse referencing")
        self.assertIsNone(coverage.taxonomic_system, "Taxonomic System from nowhere")
        self.assertIsNone(coverage.general_coverage, "General coverage from nowhere")
        self.assertEqual(1, len(coverage.classification), "Error on parsing classification")
        self.assertEqual("Kingdom", coverage.classification[0].rank_name, "Error on parsing taxonomic rank name")
        self.assertEqual("Viridiplantae", coverage.classification[0].rank_value, "Error on parsing taxonomic rank value")
        self.assertEqual(1, len(coverage.classification[0].common_name), "Error on parsing xml_common name")
        self.assertEqual("Green Plants", coverage.classification[0].common_name[0], "Error on parsing xml_common name")
        self.assertEqual(1, len(coverage.classification[0].taxon_id), "Error on parsing taxon id")
        self.assertEqual(
            33090,
            coverage.classification[0].taxon_id[0],
            "Error on parsing taxon id"
        )
        self.assertEqual(
            "https://www.ncbi.nlm.nih.gov/taxonomy",
            coverage.classification[0].taxon_id[0].provider,
            "Error on parsing taxon id"
        )
        self.assertEqual(
            0, len(coverage.classification[0].classification),
            "Taxonomic classification from nowhere"
        )
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")

    def test_parse_referencing(self):
        text_xml = """
<taxonomicCoverage scope="system">
    <references system="http://gbif.org">56</references>
</taxonomicCoverage>
        """
        coverage = TaxonomicCoverage.from_string(text_xml)
        self.assertEqual("56", coverage.id, "Error on parsing id")
        self.assertEqual("http://gbif.org", coverage.references.system, "Error on parsing system of references")
        self.assertTrue(coverage.referencing, "Error on parse referencing")
        self.assertIsNone(coverage.taxonomic_system, "Taxonomic System from nowhere")
        self.assertIsNone(coverage.general_coverage, "General coverage from nowhere")
        self.assertEqual(0, len(coverage.classification), "Classification from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), coverage.to_element(), "Error on to element")


if __name__ == '__main__':
    unittest.main()



import datetime as dt
from typing import List, Union

from datetime_interval import Interval

from dwca.terms import Field


class IdentificationID(Field):
    """
    An identifier for the Identification (the body of information associated with the assignment of a scientific name).
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identificationID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimIdentification(Field):
    """
    A string representing the taxonomic identification as it appeared in the original record.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimIdentification"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class IdentificationQualifier(Field):
    """
    A brief phrase or a standard term ("cf.", "aff.") to express the determiner's doubts about the Identification.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identificationQualifier"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class TypeStatus(Field):
    """
    A list of nomenclatural types (type status, typified scientific name, publication) applied to the subject.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/typeStatus"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class IdentifiedBy(Field):
    """
    A list of names of people, groups, or organizations who assigned the Taxon to the subject.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identifiedBy"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class IdentifiedByID(Field):
    """
    A list of the globally unique identifier for responsible for assigning the Taxon to the subject.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identifiedByID"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DateIdentified(Field):
    """
    The date on which the subject was determined as representing the Taxon.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/dateIdentified"
    TYPE = Union[dt.datetime, Interval]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class IdentificationReferences(Field):
    """
    A list of references (publication, global unique identifier, URI) used in the Identification.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identificationReferences"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class IdentificationVerificationStatus(Field):
    """
    A categorical indicator of the extent to which the taxonomic identification has been verified to be correct.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identificationVerificationStatus"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class IdentificationRemarks(Field):
    """
    Comments or notes about the Identification.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/identificationRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



import unittest

from lxml import etree as et

from dwca.terms import HigherClassification
from test_xml.test_xml import TestXML


class TestListStringField(TestXML):
    def test_parse(self):
        higher_class_xml = """<field index="3" term="http://rs.tdwg.org/dwc/terms/higherClassification"/>"""
        decimal = HigherClassification.from_string(higher_class_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/higherClassification",
            decimal.uri,
            "Wrong URI"
        )
        self.assertEqual(3, decimal.index, "Wrong index")
        self.assertEqualTree(et.fromstring(higher_class_xml), decimal.to_element(), "Wrong generated element.")
        pass

    def test_format(self):
        higher_classification = HigherClassification(3)
        value = "Plantae | Magnoliophyta | Magnoliopsida | Canellales | Winteraceae | Drimys"
        self.assertEqual(
            6, len(higher_classification.format(value)), "Error formatting list of string"
        )
        self.assertEqual(
            ["Plantae", "Magnoliophyta", "Magnoliopsida", "Canellales", "Winteraceae", "Drimys"],
            higher_classification.format(value), "Error formatting list of string"
        )
        value = ""
        self.assertEqual(
            1, len(higher_classification.format(value)), "Error formatting list of string"
        )
        self.assertEqual(
            [""],
            higher_classification.format(value), "Error formatting list of string"
        )


if __name__ == '__main__':
    unittest.main()



import os.path
import sys
import unittest
from unittest.mock import patch

from test_dwca_classes.test_outside_class_common import TestOutsideCommon

PATH = os.path.abspath(os.path.dirname(__file__))

orig_import = __import__


def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'pandas':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


class TestOutsideNoPandas(TestOutsideCommon):
    @patch('builtins.__import__', side_effect=import_mock)
    def test_parse(self, mock_import):
        super().__test_parse__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_none(self, mock_import):
        super().__test_none__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_merge(self, mock_import):
        super().__test_merge__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_primary_key(self, mock_import):
        self.__test_set_primary_key__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_sql_table(self, mock_import):
        self.__test_sql_table__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_insert(self, mock_import):
        self.__test_insert__()


if __name__ == '__main__':
    unittest.main()



from typing import List

from dwca.terms import Field


class MaterialEntityID(Field):
    """
    An identifier for a particular instance of a MaterialEntity.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/materialEntityID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Preparations(Field):
    """
    A list of preparations and preservation methods for a MaterialEntity.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/preparations"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Disposition(Field):
    """
    The current state of a MaterialEntity with respect to a collection.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/disposition"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class VerbatimLabel(Field):
    """
    Raw label derived from human transcription.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimLabel"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class AssociatedSequences(Field):
    """
    A list of identifiers (e.g. URI) of genetic sequence information associated with the MaterialEntity.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/associatedSequences"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaterialEntityRemarks(Field):
    """
    Comments or notes about the MaterialEntity instance.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/materialEntityRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



from __future__ import annotations

from typing import Dict, Union, Any

from lxml import etree as et

from xml_common.utils import Language
from xml_common import XMLObject
from eml.types import I18nString


class EMLSection(XMLObject):
    """
    A section of related text.

    The "section" element allows for grouping related paragraphs
    of text together, with an optional title.

    Parameters
    ----------
    title : I18nString or str, optional
        The optional title of the section.
    paragraph : I18String or str, optional
        A simple paragraph of text. If section is not given, this parameter is mandatory.
    section : Section, optional
        A section of related text. If paragraph is not given, this parameter is mandatory.
    language : Language, optional
        Language of section. Default `"ENG"`
    """
    PRINCIPAL_TAG = "section"

    def __init__(
            self, title: Union[I18nString, str] = None,
            paragraph: Union[I18nString, str] = None,
            section: EMLSection = None,
            language: Language = Language.ENG,
    ) -> None:
        super().__init__()
        if title is None:
            self.__title__ = None
        else:
            self.__title__ = I18nString(title)
        if paragraph is None and section is None:
            raise ValueError("Paragraph or section must be specified")
        if paragraph is None:
            self.__paragraph__ = None
        else:
            self.__paragraph__ = I18nString(paragraph)
        self.__section__ = section
        self.__lang__ = Language.ENG
        if language is None:
            self.__lang__ = Language.ENG
        else:
            self.__lang__ = language
        return

    @property
    def title(self) -> I18nString:
        """I18nString: The optional title of the section."""
        return self.__title__

    @property
    def paragraph(self) -> I18nString:
        """I18nString: A simple paragraph of text."""
        return self.__paragraph__

    @property
    def section(self) -> EMLSection:
        """EMLSection: A section of related text."""
        return self.__section__

    @property
    def language(self) -> Language:
        """Language: Language of section."""
        return self.__lang__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLSection | None:
        """
        Generate an EMLSection from an XML element.
        
        Parameters
        ----------
        element : lxml.etree.Element
            XML Element to be parsed.
        nmap : Dict
            Namespace of the XML element.

        Returns
        -------
        EMLSection
            Parsed object
        """
        if element is None:
            return None
        subsection = None
        sect_elem = element.find("section", nmap)
        if sect_elem is not None:
            subsection = EMLSection.parse(sect_elem, nmap)
        language = None
        lang_elem = element.get(f"{{{cls.NAMESPACES['xml']}}}lang", None)
        if lang_elem is not None:
            language = Language.get_language(lang_elem)
        section = EMLSection(
            title=I18nString.parse(element.find("title", nmap), nmap),
            paragraph=I18nString.parse(element.find("para", nmap), nmap),
            section=subsection,
            language=language
        )
        section.__namespace__ = nmap
        return section

    def to_element(self) -> et.Element:
        """
        Get an XML Element from the EMLSection object

        Returns
        -------
        lxml.etree.Element
            XML Element with the information of the EMLSection instance
        """
        section = super().to_element()
        if self.title is not None:
            self.title.set_tag("title")
            section.append(self.title.to_element())
        if self.paragraph is not None:
            self.paragraph.set_tag("para")
            section.append(self.paragraph.to_element())
        if self.section is not None:
            section.append(self.section.to_element())
        return section

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EMLSection):
            return (
                self.title == other.title and
                self.paragraph == other.paragraph and
                self.section == other.section
            )
        else:
            return False



import unittest

from lxml import etree as et

from eml.resources.distribution import EMLOnline
from eml.types import EMLTextType
from test_xml.test_xml import TestXML


class TestOnline(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
        "scope": "document",
        "function": "download",
    }

    def test_parse_invalid(self):
        text_xml = """
<online>
    <onlineDescription>An invalid example</onlineDescription>
    <url function="download">http://data.org/getdata?id=98332</url>
    <connection>
        <references>1</references>
    </connection>
    <connectionDefinition>
        <references>1</references>
    </connectionDefinition>
</online>
        """
        self.assertRaises(
            ValueError,
            EMLOnline.from_string,
            text_xml
        )

    def test_parse_invalid_two_connection(self):
        text_xml = """
<online>
    <onlineDescription>An invalid example</onlineDescription>
    <connection>
        <references>1</references>
    </connection>
    <connectionDefinition>
        <references>1</references>
    </connectionDefinition>
</online>
        """
        self.assertRaises(
            ValueError,
            EMLOnline.from_string,
            text_xml
        )

    def test_parse_url(self):
        text_xml = """
<online>
    <url>http://data.org/getdata?id=98332</url>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.online_description, "Online description from nowhere")
        self.assertEqual("http://data.org/getdata?id=98332", online.url, "Error on parse url")
        self.assertEqual(EMLOnline.FunctionType.DOWNLOAD, online.url_function, "Error on parse url function")
        self.assertIsNone(online.connection, "Connection from nowhere")
        self.assertIsNone(online.connection_definition, "Connection definition from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_url_with_function(self):
        text_xml = """
<online>
    <onlineDescription>An example with url</onlineDescription>
    <url function="information">http://data.org/getdata?id=98332</url>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertEqual("http://data.org/getdata?id=98332", online.url, "Error on parse url")
        self.assertEqual(EMLOnline.FunctionType.INFORMATION, online.url_function, "Error on parse url function")
        self.assertIsNone(online.connection, "Connection from nowhere")
        self.assertIsNone(online.connection_definition, "Connection definition from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_connection_referencing(self):
        text_xml = """
<online>
    <onlineDescription>An connection example</onlineDescription>
    <connection>
        <references>1</references>
    </connection>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.url, "URL from nowhere")
        self.assertEqual(
            EMLOnline.Connection(
                _id="1",
                referencing=True,
            ),
            online.connection,
            "Connection wrong parse"
        )
        self.assertIsNone(online.connection_definition, "Connection definition from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_connection_conn_definition(self):
        text_xml = """
<online>
    <onlineDescription>An connection example</onlineDescription>
    <connection>
        <connectionDefinition>
            <references>1</references>
        </connectionDefinition>
    </connection>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.url, "URL from nowhere")
        self.assertEqual(
            EMLOnline.ConnectionDefinition(
                _id="1",
                referencing=True,
            ),
            online.connection.connection_definition,
            "Connection wrong parse"
        )
        self.assertIsNone(online.connection_definition, "Connection definition from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_connection_parameters(self):
        text_xml = """
<online>
    <onlineDescription>An connection example</onlineDescription>
    <connection>
        <parameter>
            <name>hostname</name>
            <value>nceas.ucsb.edu</value>
        </parameter>
        <parameter>
            <name>port</name>
            <value>1234</value>
        </parameter>
    </connection>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.url, "URL from nowhere")
        self.assertEqual(
            2,
            len(online.connection.parameters),
            "Connection wrong parse"
        )
        self.assertEqual(
            ("hostname", "nceas.ucsb.edu"),
            online.connection.parameters[0],
            "Connection first parameter wrong parse"
        )
        self.assertEqual(
            ("port", "1234"),
            online.connection.parameters[1],
            "Connection second parameter wrong parse"
        )
        self.assertIsNone(online.connection_definition, "Connection definition from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_connection_definition_scheme(self):
        text_xml = """
<online>
    <onlineDescription>Connection definition example</onlineDescription>
    <connectionDefinition>
        <schemeName system="http://knb.ecoinformatics.org/knb/">metacat</schemeName>
    </connectionDefinition>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.url, "URL from nowhere")
        self.assertIsNone(online.connection, "Connection from nowhere")
        self.assertEqual(
            "metacat",
            online.connection_definition.scheme_name,
            "Connection definition incorrect parse"
        )
        self.assertEqual(
            "http://knb.ecoinformatics.org/knb/",
            online.connection_definition.scheme_name.system,
            "Connection definition incorrect parse system"
        )
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_connection_definition_description(self):
        description = ('The metacat application protocol. Applications '
                       'must first log into metacat by sending an HTTP '
                       'POST request in http-url-encoded format with the '
                       'parameters action, username, and password. Action '
                       'must be set to "login". If authentication is '
                       'successful, the metacat server will respond with '
                       'a session cookie. All future requests should '
                       'include the session cookie in the HTTP header. '
                       'To retrieve an object, the client then would send '
                       'an HTTP POST in http-url-encoded format, with an '
                       'action parameter set to "get" and the docid '
                       'parameter set to the identifier for the desired '
                       'object.  The response will either be an XML '
                       'document or a multipart-form-encoded response '
                       'containing data.')
        text_xml = f"""
<online>
    <onlineDescription>Connection definition example</onlineDescription>
    <connectionDefinition>
        <description>
            <para>{description}</para>
        </description>
    </connectionDefinition>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.url, "URL from nowhere")
        self.assertIsNone(online.connection, "Connection from nowhere")
        self.assertEqual(
            EMLTextType(paragraphs=[description]),
            online.connection_definition.description,
            "Connection definition incorrect parse"
        )
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parse_connection_definition_parameter(self):
        definition = ('The fully qualified name of the internet host that '
                      'is providing the metacat service, as would be '
                      'returned by a Domain Name System (DNS) query.')
        text_xml = f"""
<online>
    <onlineDescription>Connection definition example</onlineDescription>
    <connectionDefinition>
        <parameterDefinition>
            <name>hostname</name>
            <definition>{definition}</definition>
            <defaultValue>metacat.nceas.ucsb.edu</defaultValue>
        </parameterDefinition>
    </connectionDefinition>
</online>
        """
        online = EMLOnline.from_string(text_xml)
        self.assertIsNone(online.url, "URL from nowhere")
        self.assertIsNone(online.connection, "Connection from nowhere")
        self.assertEqual(
            EMLOnline.ParameterDefinition(
                name="hostname",
                definition=definition,
                default_value="metacat.nceas.ucsb.edu",
            ),
            online.connection_definition.parameters[0],
            "Connection definition incorrect parse"
        )
        self.assertEqualTree(et.fromstring(text_xml), online.to_element(), "Error on to element")

    def test_parameter_definition_operations(self):
        definition = ('The fully qualified name of the internet host that '
                      'is providing the metacat service, as would be '
                      'returned by a Domain Name System (DNS) query.')
        param1 = EMLOnline.ParameterDefinition(
            name="hostname",
            definition=definition
        )
        param2 = EMLOnline.ParameterDefinition(
            name="port",
            definition="The port in which to connect."
        )
        self.assertNotEqual(param1, param2, "Different parameter equal")
        self.assertNotEqual(3, param2, "Different types equal")
        self.assertRaises(
            TypeError, param2.__gt__, 4
        )
        self.assertGreaterEqual(param2, param1, "hostname > port")
        self.assertRaises(
            TypeError, param2.__lt__, 4
        )
        self.assertLessEqual(param1, param2, "port < hostname")

    def test_conn_definition_equal(self):
        conn_def_1 = EMLOnline.ConnectionDefinition(_id="1", referencing=True)
        conn_def_2 = EMLOnline.ConnectionDefinition(_id="1", referencing=True)
        self.assertEqual(conn_def_1, conn_def_2, "Equal elements not equal")
        self.assertNotEqual(4, conn_def_2, "Different types equal")

    def test_conn_equal(self):
        conn_1 = EMLOnline.Connection(_id="1", referencing=True)
        conn_2 = EMLOnline.Connection(_id="1", referencing=True)
        self.assertEqual(conn_1, conn_2, "Equal elements not equal")
        self.assertNotEqual(4, conn_2, "Different types equal")

    def test_parse_none(self):
        self.assertIsNone(EMLOnline.parse(None, {}), "EML Online from nowhere")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Dict, Union

from lxml import etree as et

from xml_common import XMLObject


class EMLPhone(XMLObject):
    """
    Information about the contact's telephone.

    Parameters
    ----------
    phone : str
        The phone field describes information about the responsible party's telephone.
    phone_type : str, optional
        The type of the phone to which this number applies. Default "voice".
    """
    PRINCIPAL_TAG = "phone"

    def __init__(self, phone: str, phone_type: str = "voice") -> None:
        super().__init__()
        self.__phone__ = phone
        self.__phone_type__ = phone_type
        return

    @property
    def phone(self) -> str:
        """
        str: The phone field describes information about the responsible party's telephone.
        """
        return self.__phone__

    @property
    def phone_type(self) -> str:
        """
        str: The type of the phone to which this number applies.
        """
        return self.__phone_type__

    def __eq__(self, other: Union[EMLPhone, str]) -> bool:
        if isinstance(other, str):
            return self.phone == other
        elif isinstance(other, EMLPhone):
            return self.phone == other.phone and self.phone_type == other.phone_type
        else:
            return False

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLPhone | None:
        """
        Parses a lxml element into an EML Phone object.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        XMLObject
            Object parsed.
        """
        phone = EMLPhone(element.text, element.get("phonetype", "voice"))
        phone.__namespace__ = nmap
        return phone

    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the EML Phone object.

        Returns
        -------
        lxml.tree.Element
            Object in the Element format.
        """
        phone_elem = super().to_element()
        phone_elem.set("phonetype", self.phone_type)
        phone_elem.text = self.phone
        return phone_elem



from __future__ import annotations
from typing import Dict, List, Tuple, Any

from lxml import etree as et

from xml_common import XMLObject
from eml.types import EMLObject, Scope, ResponsibleParty


class TaxonomicCoverage(EMLObject):
    """
    Taxonomic coverage information.

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    taxonomic_system : TaxonomicSystem, optional
        Documentation of taxonomic sources, procedures, and treatments.
    general_coverage : str, optional
        A description of the range of taxa addressed in the data set or collection.
    classification : List[ClassificationSystem], optional
        Information about the range of taxa addressed in the data set or collection.
    """
    class ClassificationSystem:
        """
        Information about the classification system or authority used.

        Parameters
        ----------
        citation : EMLCitation
            Relevant literature for documenting the used classification system.
        modifications : str, optional
            A description of any modifications or exceptions made to the classification system or authority used.
        """
        def __init__(self, citation: EMLObject, modifications: str = None) -> None:
            self.__cite__ = citation
            self.__modifications__ = modifications

        @property
        def citation(self) -> EMLObject:
            """EMLCitation: Relevant literature for documenting the used classification system."""
            return self.__cite__

        @property
        def modifications(self) -> str:
            """
            str: A description of any modifications or exceptions made to the classification system or authority used.
            """
            return self.__modifications__

    class Voucher:
        """
        Information on the types of specimen, the repository, and the individuals who identified the vouchers.

        Parameters
        ----------
        specimen : str
            A word or phrase describing the type of specimen collected.
        repository : List[ResponsibleParty]
            Information about the curator or contact person and/or agency responsible for the specimens.
        """
        def __init__(self, specimen: str, repository: List[ResponsibleParty]) -> None:
            self.__specimen__ = specimen
            self.__repository__ = repository

        @property
        def specimen(self) -> str:
            """str: A word or phrase describing the type of specimen collected."""
            return self.__specimen__

        @property
        def repository(self) -> List[ResponsibleParty]:
            """List[ResponsibleParty]: Information about the responsible for the specimens."""
            return self.__repository__

    class TaxonID:
        """
        Element holds an ID and ID-provider for this taxon.

        Parameters
        ----------
        _id: int
            The identifier for this taxon from an authority.
        provider: str
            The taxonomic authority from which the taxonId can be retrieved, defined as the namespace URI.
        """
        def __init__(self, _id: int, provider: str) -> None:
            self.__id__ = _id
            self.__provider__ = provider
            return

        @property
        def id(self) -> int:
            """int: The identifier for this taxon from an authority."""
            return self.__id__

        @property
        def provider(self) -> str:
            """str: The taxonomic authority from which the taxonId can be retrieved, defined as the namespace URI."""
            return self.__provider__

        def __eq__(self, other: Any) -> bool:
            if isinstance(other, int):
                return self.id == other
            elif isinstance(other, TaxonomicCoverage.TaxonID):
                return self.id == other.id and self.provider == other.provider
            else:
                return False

    class TaxonomicSystem(XMLObject):
        """
        Documentation of taxonomic sources, procedures, and treatments.

        Parameters
        ----------
        procedures : str
            Description of the methods used for the taxonomic identification.
        classification_system : List[EMLCitation]
             Relevant literature for documenting the used classification system.
        identifier_name : List[ResponsibleParty]
             Information about the individual(s) responsible for the identification(s) of the specimens or sightings.
        identification_reference : List[EMLCitation], optional
            Information on any non-authoritative materials.
        completeness : str, optional
            Taxonomic completeness.
        vouchers : List[Tuple[str, List[ResponsibleParty]]], optional
            Information on the types of specimen, the repository, and the individuals who identified the vouchers.
        modifications : List[str], optional
            A description of any modifications or exceptions made to the classification system or authority used.
        """
        PRINCIPAL_TAG = "taxonomicSystem"

        def __init__(
                self, procedures: str,
                classification_system: List,
                identifier_name: List[ResponsibleParty],
                identification_reference: List = None,
                completeness: str = None,
                vouchers: List[Tuple[str, List[ResponsibleParty]]] = None,
                modifications: List[str] = None,
        ) -> None:
            super().__init__()
            self.__procedures__ = procedures
            self.__class_system__ = list()
            if len(classification_system) == 0:
                raise ValueError("At least one classification system must be given")
            for i, class_sys in enumerate(classification_system):
                try:
                    class_mod = modifications[i]
                except IndexError:
                    class_mod = None
                except TypeError:
                    class_mod = None
                self.__class_system__.append(
                    TaxonomicCoverage.ClassificationSystem(class_sys, modifications=class_mod)
                )
            self.__id_ref__ = list()
            if identification_reference is not None:
                self.__id_ref__.extend(identification_reference)
            self.__id_name__ = list()
            if len(identifier_name) == 0:
                raise ValueError("At least one identifier name must be given")
            self.__id_name__ = identifier_name
            self.__completeness__ = completeness
            self.__vouchers__ = list()
            if vouchers is not None:
                for specimen, repository in vouchers:
                    self.__vouchers__.append(
                        TaxonomicCoverage.Voucher(specimen, repository)
                    )
            return

        @property
        def procedures(self) -> str:
            """str: Description of the methods used for the taxonomic identification."""
            return self.__procedures__

        @property
        def classification_system(self) -> List[TaxonomicCoverage.ClassificationSystem]:
            """List[ClassificationSystem]: Information about the classification system or authority used."""
            return self.__class_system__

        @property
        def identification_reference(self) -> List:
            """List[EMLCitation]: Information on any non-authoritative materials."""
            return self.__id_ref__

        @property
        def identifier_name(self) -> List[ResponsibleParty]:
            """
            List[ResponsibleParty]: Information about the individual(s) responsible for the identification(s).
            """
            return self.__id_name__

        @property
        def completeness(self) -> str:
            """
            str: Taxonomic completeness.
            """
            return self.__completeness__

        @property
        def vouchers(self) -> List[TaxonomicCoverage.Voucher]:
            """List[Voucher]: Information on the types of specimen and the repository."""
            return self.__vouchers__

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> TaxonomicCoverage.TaxonomicSystem | None:
            """
            Generate a TaxonomicSystem instance from an XML element instance.

            Parameters
            ----------
            element : lxml.etree.Element
                XML element to parse.
            nmap : Dict
                Namespace.

            Returns
            -------
            TaxonomicSystem
                Instance with the same information as the XML element.
            """
            if element is None:
                return None
            procedures_elem = element.find("taxonomicProcedures", nmap)
            classification_system = list()
            modifications = list()
            from eml.resources import EMLCitation
            for class_sys in element.findall("classificationSystem", nmap):
                citation_elem = class_sys.find("classificationSystemCitation", nmap)
                classification_system.append(EMLCitation.parse(citation_elem, nmap))
                modification_elem = class_sys.find("classificationSystemModifications", nmap)
                if modification_elem is not None:
                    modifications.append(modification_elem.text)
                else:
                    modifications.append(None)
            identification_ref = list()
            for id_ref in element.findall("identificationReference", nmap):
                identification_ref.append(EMLCitation.parse(id_ref, nmap))
            identification_name = list()
            for id_name in element.findall("identifierName", nmap):
                identification_name.append(ResponsibleParty.parse(id_name, nmap))
            completeness = None
            completeness_elem = element.find("taxonomicCompleteness", nmap)
            if completeness_elem is not None:
                completeness = completeness_elem.text if completeness_elem.text is not None else ""
            vouchers = list()
            for voucher_elem in element.findall("vouchers", nmap):
                specimen = voucher_elem.find("specimen", nmap)
                originator = list()
                for ori_elem in voucher_elem.find("repository", nmap).findall("originator", nmap):
                    originator.append(ResponsibleParty.parse(ori_elem, nmap))
                vouchers.append((
                    specimen.text if specimen.text is not None else "",
                    originator
                ))
            taxonomic_system = TaxonomicCoverage.TaxonomicSystem(
                procedures=procedures_elem.text if procedures_elem.text is not None else "",
                classification_system=classification_system,
                identification_reference=identification_ref,
                identifier_name=identification_name,
                completeness=completeness,
                vouchers=vouchers,
                modifications=modifications,
            )
            taxonomic_system.__namespace__ = nmap
            return taxonomic_system

        def to_element(self) -> et.Element:
            """
            Generate an XML instance using this object.

            Returns
            -------
            lxml.etree.Element
                XML element instance.
            """
            element = super().to_element()
            for class_sys in self.classification_system:
                class_elem = self.object_to_element("classificationSystem")
                cite_elem = class_sys.citation.to_element()
                cite_elem.tag = "classificationSystemCitation"
                class_elem.append(cite_elem)
                if class_sys.modifications is not None:
                    mod_elem = self.object_to_element("classificationSystemModifications")
                    mod_elem.text = class_sys.modifications
                    class_elem.append(mod_elem)
                element.append(class_elem)
            for id_ref in self.identification_reference:
                id_elem = id_ref.to_element()
                id_elem.tag = "identificationReference"
                element.append(id_elem)
            for id_name in self.identifier_name:
                id_name.set_tag("identifierName")
                element.append(id_name.to_element())
            procedures_elem = self.object_to_element("taxonomicProcedures")
            procedures_elem.text = self.procedures
            element.append(procedures_elem)
            if self.completeness:
                completeness_elem = self.object_to_element("taxonomicCompleteness")
                completeness_elem.text = self.completeness
                element.append(completeness_elem)
            for voucher in self.vouchers:
                voucher_elem = self.object_to_element("vouchers")
                specimen_elem = self.object_to_element("specimen")
                specimen_elem.text = voucher.specimen
                voucher_elem.append(specimen_elem)
                repository_elem = self.object_to_element("repository")
                for originator in voucher.repository:
                    originator.set_tag("originator")
                    repository_elem.append(originator.to_element())
                voucher_elem.append(repository_elem)
                element.append(voucher_elem)
            return element

    class TaxonomicClassification(XMLObject):
        """
        Information about the range of taxa addressed in the data set or collection.

        Parameters
        ----------
        _id : str, optional
            A unique identifier for this additional metadata that can be used to reference it elsewhere.
        rank_name : str, optional
            The name of the taxonomic rank for which the Taxon rank value is provided.
        rank_value : str, optional
            The taxonomic rank name being described.
        common_name: List[str], optional
            Specification of applicable xml_common names.
        taxon_id: List[Tuple[int, str]], optional
            Element holds an ID and ID-provider for this taxon.
        classification: List[TaxonClassification], optional
            Taxonomic Classification field is self-referencing to allow for an arbitrary depth of rank, down to species.
        """
        PRINCIPAL_TAG = "taxonomicClassification"

        def __init__(
                self, _id: str = None,
                rank_name: str = None,
                rank_value: str = None,
                common_name: List[str] = None,
                taxon_id: List[Tuple[int, str]] = None,
                classification: List[TaxonomicCoverage.TaxonomicClassification] = None,
        ) -> None:
            super().__init__()
            self.__id__ = _id
            self.__name__ = rank_name
            self.__value__ = rank_value
            self.__common__ = list()
            if common_name is not None:
                self.__common__.extend(common_name)
            self.__id_taxon__ = list()
            if taxon_id is not None:
                for t_id, provider in taxon_id:
                    self.__id_taxon__.append(
                        TaxonomicCoverage.TaxonID(t_id, provider)
                    )
            self.__classification__ = list()
            if classification is not None:
                self.__classification__.extend(classification)
            return

        @property
        def id(self) -> str:
            """str: A unique identifier for this additional metadata that can be used to reference it elsewhere."""
            return self.__id__

        @property
        def rank_name(self) -> str:
            """str: The name of the taxonomic rank for which the Taxon rank value is provided."""
            return self.__name__

        @property
        def rank_value(self) -> str:
            """str: The taxonomic rank name being described."""
            return self.__value__

        @property
        def common_name(self) -> List[str]:
            """List[str]: Specification of applicable xml_common names."""
            return self.__common__

        @property
        def taxon_id(self) -> List[TaxonomicCoverage.TaxonID]:
            """List[TaxonID]: Element holds an ID and ID-provider for this taxon."""
            return self.__id_taxon__

        @property
        def classification(self) -> List[TaxonomicCoverage.TaxonomicClassification]:
            """List[TaxonomicClassification]: Taxonomic Classification field self-referenced."""
            return self.__classification__

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> TaxonomicCoverage.TaxonomicClassification | None:
            """
            Generate a TaxonomicClassification instance from an XML element.

            Parameters
            ----------
            element : lxml.etree.Element
                XML Element to be parsed.
            nmap : Dict
                Namespace

            Returns
            -------
            TaxonomicClassification
                An instance with the XML element information.
            """
            if element is None:
                return None
            rank_name = None
            name_elem = element.find("taxonRankName", nmap)
            if name_elem is not None:
                rank_name = name_elem.text if name_elem.text is not None else ""
            rank_value = None
            value_elem = element.find("taxonRankValue", nmap)
            if value_elem is not None:
                rank_value = value_elem.text if value_elem.text is not None else ""
            common_names = list()
            for common_elem in element.findall("commonName", nmap):
                common_names.append(common_elem.text if common_elem.text is not None else "")
            taxon_id = list()
            for taxon_elem in element.findall("taxonId", nmap):
                taxon_id.append((
                    int(taxon_elem.text), taxon_elem.get("provider")
                ))
            classifications = list()
            for classification_elem in element.findall("taxonomicClassification"):
                classifications.append(TaxonomicCoverage.TaxonomicClassification.parse(classification_elem, nmap))
            return TaxonomicCoverage.TaxonomicClassification(
                _id=element.get("id", None),
                rank_name=rank_name,
                rank_value=rank_value,
                common_name=common_names,
                taxon_id=taxon_id,
                classification=classifications
            )

        def to_element(self) -> et.Element:
            """
            Generate an XML element instance with the information of this object.

            Returns
            -------
            lxml.etree.Element
                XML element instance.
            """
            element = super().to_element()
            if self.id is not None:
                element.set("id", self.id)
            if self.rank_name is not None:
                name_elem = self.object_to_element("taxonRankName")
                name_elem.text = self.rank_name
                element.append(name_elem)
            if self.rank_value is not None:
                value_elem = self.object_to_element("taxonRankValue")
                value_elem.text = self.rank_value
                element.append(value_elem)
            for common_name in self.common_name:
                common_elem = self.object_to_element("commonName")
                common_elem.text = common_name
                element.append(common_elem)
            for taxon_id in self.taxon_id:
                taxon_elem = self.object_to_element("taxonId")
                taxon_elem.text = str(taxon_id.id)
                taxon_elem.set("provider", taxon_id.provider)
                element.append(taxon_elem)
            for classification in self.classification:
                element.append(classification.to_element())
            return element

    PRINCIPAL_TAG = "taxonomicCoverage"

    def __init__(
            self, _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            taxonomic_system: TaxonomicCoverage.TaxonomicSystem = None,
            general_coverage: str = None,
            classification: List[TaxonomicCoverage.TaxonomicClassification] = None,
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        self.__taxa_system__ = taxonomic_system
        self.__general__ = general_coverage
        self.__classification__ = list()
        if self.referencing:
            return
        if classification is None or len(classification) == 0:
            raise ValueError("At least one taxonomic classification must be given")
        self.__classification__.extend(classification)
        return

    @property
    def taxonomic_system(self) -> TaxonomicCoverage.TaxonomicSystem:
        """TaxonomicSystem: Documentation of taxonomic sources, procedures, and treatments."""
        return self.__taxa_system__

    @property
    def general_coverage(self) -> str:
        """str: A description of the range of taxa addressed in the data set or collection."""
        return self.__general__

    @property
    def classification(self) -> List[TaxonomicCoverage.TaxonomicClassification]:
        """TaxonomicClassification: Information about the range of taxa addressed in the data set or collection."""
        return self.__classification__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLObject:
        """
        Generate a Taxonomic Coverage referencing another Taxonomic Coverage.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        TaxonomicCoverage
            Taxonomic Coverage object parsed that reference another.
        """
        references = element.find("references", nmap)
        return TaxonomicCoverage(
            _id=references.text if references.text is not None else "",
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> TaxonomicCoverage:
        """
        Generate a Taxonomic Coverage instance that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        TaxonomicCoverage
            Object parsed.
        """
        taxon_system = None
        taxon_system_elem = element.find("taxonomicSystem", nmap)
        if taxon_system_elem is not None:
            taxon_system = TaxonomicCoverage.TaxonomicSystem.parse(taxon_system_elem, nmap)
        general_coverage = None
        general_coverage_elem = element.find("generalTaxonomicCoverage", nmap)
        if general_coverage_elem is not None:
            general_coverage = general_coverage_elem.text if general_coverage_elem.text is not None else ""
        classification = list()
        for taxon_class in element.findall("taxonomicClassification", nmap):
            classification.append(
                TaxonomicCoverage.TaxonomicClassification.parse(
                    taxon_class, nmap
                )
            )
        return TaxonomicCoverage(
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            taxonomic_system=taxon_system,
            general_coverage=general_coverage,
            classification=classification,
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance using object information.

        Returns
        -------
        lxml.etree.Element
            XML element instance.
        """
        element = super().to_element()
        element = self._to_element_(element)
        if self.referencing:
            return element
        if self.taxonomic_system is not None:
            element.append(self.taxonomic_system.to_element())
        if self.general_coverage is not None:
            general_elem = self.object_to_element("generalTaxonomicCoverage")
            general_elem.text = self.general_coverage
            element.append(general_elem)
        for classification in self.classification:
            element.append(classification.to_element())
        return element



import os.path
import sys
import unittest
from unittest.mock import patch

import pandas as pd
from lxml import etree as et

from dwca.classes import Taxon
from dwca.terms import DWCLanguage
from test_dwca_classes.test_taxon_common import TestTaxonCommon
from xml_common.utils import Language

PATH = os.path.abspath(os.path.dirname(__file__))

orig_import = __import__


def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'polars':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


def import_mock_fuzzy(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'rapidfuzz':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


class TestTaxonNoPolars(TestTaxonCommon):
    @patch('builtins.__import__', side_effect=import_mock)
    def test_add_field(self, mock_import):
        super().__test_add_field__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_add_field_incorrect_index(self, mock_import):
        super().__test_add_field_incorrect_index__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_add_field_data(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        self.assertEqual(47, len(df.columns), "Wrong number of initial columns")
        self.assertRaisesRegex(
            AttributeError, "language",
            getattr,
            self.taxon.__entries__[0],
            "language"
        )
        self.taxon.add_field(DWCLanguage(index=47, default=Language.ENG))
        self.assertEqual(
            Language.ENG,
            self.taxon.__entries__[0].language,
            "Language not write on all entries"
        )
        df = self.taxon.pandas
        self.assertEqual(48, len(df.columns), "Wrong number of columns after set")
        self.assertEqual(48, len(self.taxon.fields), "Fields not set")
        self.text = self.text.replace(
            """<field index="46" default="EXAMPLE" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>""",
            """<field index="46" default="EXAMPLE" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>
    <field index="47" default="eng" term="http://purl.org/dc/elements/1.1/language"/>""")
        self.assertEqual(Language.ENG, self.taxon.pandas.loc[7, "language"], "Wrong set default on pandas.")
        self.assertEqual(Language.ENG, self.taxon.__entries__[7].language, "Wrong set default on entries.")
        self.assertEqualTree(et.fromstring(self.text), self.taxon.to_element(), "Error on element conversion with new field")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_parse(self, mock_import):
        super().__test_parse__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_kingdom_exception(self, mock_import):
        super().__test_filter_kingdom_exception__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_parents(self, mock_import):
        super().__test_get_parents__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_incorrect_parent(self, mock_import):
        self.__test_get_incorrect_parent__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_synonyms(self, mock_import):
        self.__test_get_synonyms__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_synonyms_names(self, mock_import):
        super().__test_get_synonyms_names__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_incorrect_synonyms(self, mock_import):
        self.__test_get_incorrect_synonyms__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_kingdom(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        kingdom_summary = df.groupby("kingdom").size().to_dict()
        print(f"Kingdom Summary: {kingdom_summary}", file=sys.stderr)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_kingdom(kingdom_summary.keys())
        self.assertEqual(length_df, len(self.taxon), "Filter by all kingdoms.")
        first_kingdom = list(kingdom_summary.keys())[0]
        self.taxon.filter_by_kingdom([first_kingdom])
        self.assertEqual(kingdom_summary[first_kingdom], len(self.taxon), "Filter by first kingdom.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_phylum(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        phylum_summary = df.groupby(["kingdom", "phylum"]).size().to_dict()
        kingdom_list = list()
        phylum_list = list()
        for candid in phylum_summary.keys():
            if candid[1] != "":
                phylum_list.append(candid[1])
            else:
                kingdom_list.append(candid[0])
        print(f"Phylum Summary: {phylum_summary}", file=sys.stderr)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_phylum(phylum_list)
        self.assertEqual(length_df, len(self.taxon), "Filter by all phylum.")
        self.taxon.filter_by_phylum([phylum_list[0]])
        self.assertEqual(
            phylum_summary[(kingdom_list[0], "")] + phylum_summary[(kingdom_list[0], phylum_list[0])],
            len(self.taxon),
            f"Incorrect filter by {phylum_list[0]}."
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_class(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        class_summary = df.groupby(["kingdom", "phylum", "class"]).size().to_dict()
        kingdom_list = list()
        phylum_list = list()
        class_list = list()
        for candid in class_summary.keys():
            if candid[2] != "":
                class_list.append(candid[2])
            else:
                if candid[1] != "":
                    phylum_list.append(candid[1])
                else:
                    kingdom_list.append(candid[0])
        print(f"Class Summary: {class_summary}", file=sys.stderr)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_class(class_list)
        self.assertEqual(length_df, len(self.taxon), "Filter by all classes.")
        self.taxon.filter_by_class([class_list[2]])
        self.assertEqual(
            class_summary[(kingdom_list[0], "", "")] +
            class_summary[(kingdom_list[0], phylum_list[0], "")] +
            class_summary[(kingdom_list[0], phylum_list[0], class_list[1])] +
            class_summary[(kingdom_list[0], phylum_list[0], class_list[2])],
            len(self.taxon),
            f"Incorrect filter by {class_list[2]}."
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_order(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        orders = list(filter(lambda x: x != "", pd.unique(df["order"])))
        order_names = ", ".join([f'"{order}"' for order in orders])
        print(f"Orders found: {order_names}", file=sys.stderr)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_order(orders)
        self.assertEqual(length_df, len(self.taxon), "Filter by all orders.")
        accepted = df[df["scientificName"] == "Rbaliuycliu"]["acceptedNameUsage"].iloc[0]
        with_order = sum(df["order"].isin(["Rbaliuycliu", "Xlhofowltm", accepted]))
        taxa_id = df[df["scientificName"].isin(["Rbaliuycliu", "Xlhofowltm", accepted])]["taxonID"]
        parents = self.taxon.get_parents(taxa_id)
        parents.update(self.taxon.all_synonyms(parents))
        parents = sum(df["taxonID"].isin(parents))
        print("Filtering by orders: Rbaliuycliu, Xlhofowltm", file=sys.stderr)
        print(f"Accepted orders: Rbaliuycliu, {accepted}", file=sys.stderr)
        self.taxon.filter_by_order(["Rbaliuycliu", "Xlhofowltm"])
        print(f"Expected: with order {with_order} and parents {parents}", file=sys.stderr)
        self.assertEqual(with_order + parents, len(self.taxon), "Incorrect filter.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_family(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        families = list(filter(lambda x: x != "", pd.unique(df["family"])))
        family_names = ", ".join([f'"{family}"' for family in families])
        print(f"Families found: {family_names}", file=sys.stderr)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_family(families)
        self.assertEqual(length_df, len(self.taxon), "Filter by all families.")
        accepted = df[df["scientificName"] == "Dxaougdpy"]["acceptedNameUsage"].iloc[0]
        with_family = sum(df["family"].isin(["Dxaougdpy", "Kvtgqwfmy", accepted]))
        taxa_id = df[df["scientificName"].isin(["Dxaougdpy", "Kvtgqwfmy", accepted])]["taxonID"]
        parents = self.taxon.get_parents(taxa_id)
        parents.update(self.taxon.all_synonyms(parents))
        parents = sum(df["taxonID"].isin(parents))
        print("Filtering by family: Dxaougdpy, Kvtgqwfmy", file=sys.stderr)
        print(f"Accepted families: Kvtgqwfmy, {accepted}", file=sys.stderr)
        self.taxon.filter_by_family(["Dxaougdpy", "Kvtgqwfmy"])
        print(f"Expected: with family {with_family} and parents {parents}", file=sys.stderr)
        self.assertEqual(with_family + parents, len(self.taxon), "Incorrect filter.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_genus(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        genera = list(filter(lambda x: x != "", pd.unique(df["genus"])))
        genus_names = ", ".join([f'"{genus}"' for genus in genera])
        print(f"Genera found: {genus_names}", file=sys.stderr)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_genus(genera)
        self.assertEqual(length_df, len(self.taxon), "Filter by all genera.")
        accepted = df[df["scientificName"] == "Hbqrtfopjuh"]["acceptedNameUsage"].iloc[0]
        with_genus = sum(df["genus"].isin(["Hbqrtfopjuh", "Vbpjkmfqd", accepted]))
        taxa_id = df[df["scientificName"].isin(["Hbqrtfopjuh", "Vbpjkmfqd", accepted])]["taxonID"]
        parents = self.taxon.get_parents(taxa_id)
        parents.update(self.taxon.all_synonyms(parents))
        parents = sum(df["taxonID"].isin(parents))
        print("Filtering by genus: Hbqrtfopjuh, Vbpjkmfqd", file=sys.stderr)
        print(f"Accepted genera: Vbpjkmfqd, {accepted}", file=sys.stderr)
        self.taxon.filter_by_genus(["Hbqrtfopjuh", "Vbpjkmfqd"])
        print(f"Expected: with genus {with_genus} and parents {parents}", file=sys.stderr)
        self.assertEqual(with_genus + parents, len(self.taxon), "Incorrect filter.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_species(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        length_df = len(df)
        self.assertEqual(length_df, len(self.taxon), "Original length of dataframe.")
        species_synonym = "Hqmjhacvb (Azvoxtzhwueu) gupfjmnf"
        variety = "Rtfkiaicpdng obzninluz var. cebwyzcqoy Glhnskwn"
        cultivar = "Rtfkiaicpdng (Abifwvxqn) gurqtwpof f. prczacpvtdtu 'xzhgezqpaorp'"
        synonyms = set()
        species_row = df[df["scientificName"] == species_synonym].iloc[0]
        accepted = species_row["acceptedNameUsageID"]
        synonyms.add(accepted)
        synonyms.add(species_row["taxonID"])
        first_parents = self.taxon.get_parents([accepted])
        print(f"For {species_synonym} expected {len(first_parents)} parents, accepted and synonym", file=sys.stderr)
        variety_id = df[df["scientificName"] == variety]["taxonID"].iloc[0]
        second_parents = self.taxon.get_parents([variety_id])
        synonyms.add(variety_id)
        print(f"For {variety} expected {len(second_parents)} parents and self", file=sys.stderr)
        cultivar_id = df[df["scientificName"] == cultivar]["taxonID"].iloc[0]
        third_parents = self.taxon.get_parents([cultivar_id])
        synonyms.add(cultivar_id)
        print(f"For {cultivar} expected {len(third_parents)} parents", file=sys.stderr)
        synonyms.update(self.taxon.all_synonyms(
            list(synonyms.copy()) +
            list(first_parents) +
            list(second_parents) +
            list(third_parents)
        ))
        other_synonyms = synonyms.difference({species_synonym, variety, cultivar})
        other_synonyms = other_synonyms.difference(set(first_parents))
        other_synonyms = other_synonyms.difference(set(second_parents))
        other_synonyms = other_synonyms.difference(set(third_parents))
        print(f"Synonyms found: {len(other_synonyms)}", file=sys.stderr)
        print(f"Total {len(synonyms)}", file=sys.stderr)
        self.taxon.filter_by_species([species_synonym, variety, cultivar])
        self.assertEqual(
            len(synonyms),
            len(self.taxon),
            "Filter by all genera."
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_phylum_exception(self, mock_import):
        super().__test_filter_phylum_exception__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_none(self, mock_import):
        super().__test_none__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_merge(self, mock_import):
        super().__test_merge__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_missing_field(self, mock_import):
        self.assertRaises(
            AssertionError,
            Taxon.from_string,
            """
<core>
    <files>
      <location>missing.txt</location>
    </files>
    <id index="0"/>
    <field index="0" term=""/>
    <field index="1" term=""/>
    <field index="3" term=""/>
</core>
            """
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_field_twice(self, mock_import):
        self.assertRaises(
            AssertionError,
            Taxon.from_string,
            """
<core>
    <files>
      <location>missing.txt</location>
    </files>
    <id index="0"/>
    <field index="0" term=""/>
    <field index="1" term=""/>
    <field index="3" term=""/>
    <field index="2" term=""/>
    <field index="3" term=""/>
    <field index="4" term=""/>
</core>
            """
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_parse_invalid(self, mock_import):
        self.assertRaises(
            AssertionError,
            Taxon.from_string,
            '<extension />'
        )
        self.assertRaises(
            KeyError,
            Taxon.from_string,
            """
<core-id>
    <field index="0" term=""/>
</core-id>
            """
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_pandas(self, mock_import):
        self.read_pandas()
        df = self.taxon.pandas
        self.assertEqual(163460, len(self.taxon), "Data incorrectly read.")
        self.assertEqual(163460, len(df), "DataFrame incorrectly read.")
        df = df[0:100]
        self.taxon.pandas = df
        self.assertEqual(100, len(df), "DataFrame incorrectly set.")
        self.assertEqual(100, len(self.taxon), "Data incorrectly set.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_write_file(self, mock_import):
        self.read_pandas()
        plain_text = self.taxon.write_file()
        self.maxDiff = None
        with open(os.path.join(PATH, os.pardir, "example_data", "taxon.txt"), "r", encoding="utf-8") as file:
            for expected, actual in zip(file.read().split("\n"), plain_text.split("\n")):
                self.assertEqual(
                    expected,
                    actual,
                    "File written incorrectly."
                )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_read_lazy(self, mock_import):
        self.assertRaises(
            ImportError,
            self.read_pandas,
            lazy=True,
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_no_polars(self, mock_import):
        self.read_pandas()
        with self.assertRaisesRegex(ImportError, "Install polars to use this feature"):
            var = self.taxon.polars

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_core_field(self, mock_import):
        self.__test_set_core_field__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_primary_key(self, mock_import):
        self.__test_set_primary_key__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_sql_table(self, mock_import):
        self.__test_sql_table__()

    @patch('builtins.__import__', side_effect=import_mock_fuzzy)
    def test_fuzzy_exception(self, mock_import):
        self.__test_fuzzy_exception__()


if __name__ == '__main__':
    unittest.main()
